# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from inspect import getfullargspec
from typing import (Callable, Dict, Optional, TYPE_CHECKING, cast)

from ..core.target import Target
from ..core.memory_map import (FlashRegion, MemoryType, RamRegion, DeviceRegion, MemoryMap)
from ..core.soc_target import SoCTarget
from ..core import exceptions
from . import (dap, discovery)
from ..debug.svd.loader import SVDLoader
from ..utility.sequencer import CallSequence
from ..target.pack.flm_region_builder import FlmFlashRegionBuilder

if TYPE_CHECKING:
    from ..core.session import Session
    from ..core.memory_map import MemoryMap
    from .ap import (APAddressBase, AccessPort)
    from ..debug.svd.model import SVDDevice

LOG = logging.getLogger(__name__)

class CoreSightTarget(SoCTarget):
    """@brief Represents an SoC that uses CoreSight debug infrastructure.

    This class adds Arm CoreSight-specific discovery and initialization code to SoCTarget.
    """

    def __init__(self, session: "Session", memory_map: Optional["MemoryMap"] = None) -> None:
        # Supply a default memory map.
        if (memory_map is None) or (memory_map.region_count == 0):
            memory_map = self._create_default_cortex_m_memory_map()
            LOG.debug("Using default Cortex-M memory map (no memory map supplied)")

        super().__init__(session, memory_map)
        assert session.probe
        self.dp = dap.DebugPort(session.probe, self)
        self._svd_load_thread: Optional[SVDLoader] = None
        self._irq_table: Optional[Dict[int, str]] = None
        self._discoverer: Optional[Callable] = None

        self.session.context_state.is_performing_pre_reset = False

    @property
    def aps(self) -> Dict["APAddressBase", "AccessPort"]:
        return self.dp.aps

    @property
    def svd_device(self) -> Optional["SVDDevice"]:
        """@brief Waits for SVD file to complete loading before returning."""
        if not self._svd_device and self._svd_load_thread:
            LOG.debug("Waiting for SVD load to complete")
            self._svd_device = self._svd_load_thread.device
        return self._svd_device

    def _create_default_cortex_m_memory_map(self) -> MemoryMap:
        """@brief Create a MemoryMap for the Cortex-M system address map."""
        return MemoryMap(
                RamRegion(name="Code",          start=0x00000000, length=0x20000000, access='rwx'),
                RamRegion(name="SRAM",          start=0x20000000, length=0x20000000, access='rwx'),
                DeviceRegion(name="Peripheral", start=0x40000000, length=0x20000000, access='rw'),
                RamRegion(name="RAM1",          start=0x60000000, length=0x20000000, access='rwx'),
                RamRegion(name="RAM2",          start=0x80000000, length=0x20000000, access='rwx'),
                DeviceRegion(name="Device1",    start=0xA0000000, length=0x20000000, access='rw'),
                DeviceRegion(name="Device2",    start=0xC0000000, length=0x20000000, access='rw'),
                DeviceRegion(name="PPB",        start=0xE0000000, length=0x20000000, access='rw'),
                )

    def load_svd(self) -> None:
        def svd_load_completed_cb(svdDevice):
            self._svd_device = svdDevice
            self._svd_load_thread = None

        if not self._svd_device and self._svd_location:
            # Spawn thread to load SVD in background.
            self._svd_load_thread = SVDLoader(self._svd_location, svd_load_completed_cb)
            self._svd_load_thread.load()

    def create_init_sequence(self) -> CallSequence:
        seq = CallSequence(
            ('load_svd',            self.load_svd),
            ('pre_connect',         self.pre_connect),
            ('dp_init',             self.dp.create_connect_sequence),
            ('unlock_device',       self.unlock_device),
            ('create_discoverer',   self.create_discoverer),
            ('discovery',           lambda : self._discoverer.discover() if self._discoverer else None),
            ('check_for_cores',     self.check_for_cores),
            ('halt_on_connect',     self.perform_halt_on_connect),
            ('post_connect',        self.post_connect),
            ('post_connect_hook',   self.post_connect_hook),
            ('create_flash',        self.create_flash),
            ('notify',              lambda : self.session.notify(Target.Event.POST_CONNECT, self))
            )

        return seq

    def init(self) -> None:
        """@brief CoreSight specific target init.

        Connects this object's delegates, including a debug sequence delegate, to the DP.
        """
        # Set delegates on the DP.
        self.dp.delegate = self.delegate
        if self.debug_sequence_delegate:
            self.dp.debug_sequence_delegate = self.debug_sequence_delegate

        super().init()

    def disconnect(self, resume: bool = True) -> None:
        """@brief Disconnect from the target.

        Same as SoCTarget.disconnect(), except that it asks the DebugPort to power down.
        """
        self.session.notify(Target.Event.PRE_DISCONNECT, self)
        self.call_delegate('will_disconnect', target=self, resume=resume)
        for core in self.cores.values():
            core.disconnect(resume)
        # Only disconnect the DP if resuming; if not resuming we need to keep debug powered up so
        # the core can stay halted.
        if resume:
            self.dp.disconnect()
        self.call_delegate('did_disconnect', target=self, resume=resume)

    @property
    def primary_core_pname(self) -> str:
        """@brief Returns the pname for the `primary_core` option.

        This property is expected to be used prior to discovery. After discovery is complete, the
        node name of the `.primary_core` property can be used.

        This property is used rarely, so is not cached.

        @exception KeyError if `primary_core` is invalid.
        @exception AssertionError The device is not DFP based.
        """
        # The `primary_core` is an index into available cores, not necessarily the same as the AP
        # address and always different in the case of ADIv6. So we must use the DFP's list of APs
        # and processors to reconstruct the core order we will find during discovery.
        assert self.debug_sequence_delegate
        pack_device = self.debug_sequence_delegate.cmsis_pack_device
        ap_map = pack_device.processors_ap_map
        primary_core = self.session.options.get('primary_core')
        for i, proc_info in enumerate(sorted(ap_map.values(), key=lambda p: p.ap_address)):
            if i == primary_core:
                return proc_info.name
        else:
            raise exceptions.Error(f"invalid 'primary_core' session option '{primary_core}' "
                           f"(valid values are {', '.join(str(i) for i, _ in enumerate(ap_map.values()))})")

    def call_pre_discovery_debug_sequence(self, sequence: str) -> bool:
        """@brief Run a debug sequence before discovery has been performed.

        The primary core's pname cannot be looked up via the `node_name` property of the core
        object because that core object doesn't yet exist at the time this method is called.
        """
        if self.debug_sequence_delegate:
            # Try to get the pname to use.
            try:
                pcore_pname = self.primary_core_pname
            except exceptions.Error as err:
                LOG.warning("%s", err)
            else:
                if self.has_debug_sequence(sequence, pname=pcore_pname):
                    self.debug_sequence_delegate.run_sequence(sequence, pname=pcore_pname)
                    return True

        # Sequence wasn't run.
        return False

    def unlock_device(self) -> None:
        """@brief Hook to unlock the debug.

        The default implementation of this hook calls the delegate `unlock_device()` method or `DebugDeviceUnlock`
        debug sequence, if they exist, checked in this order.
        """
        if self.delegate_implements('unlock_device'):
            self.call_delegate('unlock_device')
        else:
            self.call_pre_discovery_debug_sequence('DebugDeviceUnlock')

    def create_discoverer(self) -> None:
        """@brief Init task to create the discovery object.

        Instantiates the appropriate @ref pyocd.coresight.discovery.CoreSightDiscovery
        CoreSightDiscovery subclass for the target's ADI version.
        """
        self._discoverer = discovery.ADI_DISCOVERY_CLASS_MAP[self.dp.adi_version](self)

    def pre_connect(self) -> None:
        """@brief Handle some of the connect modes.

        This init task performs a connect pre-reset or asserts reset if the connect mode is
        under-reset.
        """
        mode = self.session.options.get('connect_mode')
        if mode == 'pre-reset':
            LOG.info("Performing connect pre-reset")
            try:
                # Set the state variable indicating we're running ResetHardware for pre-reset, used
                # by the debug sequence delegate's get_connection_type() method.
                self.session.context_state.is_performing_pre_reset = True
                if not self.call_pre_discovery_debug_sequence('ResetHardware'):
                    self.dp.reset()
            finally:
                self.session.context_state.is_performing_pre_reset = False
        elif mode == 'under-reset':
            LOG.info("Asserting reset prior to connect")
            self.dp.assert_reset(True)

    def perform_halt_on_connect(self) -> None:
        """@brief Halt cores.

        This init task performs a connect pre-reset or asserts reset if the connect mode is
        under-reset.
        """
        mode = self.session.options.get('connect_mode')
        if mode != 'attach':
            if mode == 'under-reset':
                LOG.debug("Setting reset catch")
            # Apply to all cores.
            for core in self.cores.values():
                try:
                    if mode == 'under-reset':
                        core.set_reset_catch(Target.ResetType.HW)
                    else:
                        core.halt()
                except exceptions.Error as err:
                    LOG.warning("Could not halt core #%d: %s", core.core_number, err,
                        exc_info=self.session.log_tracebacks)

    def post_connect(self) -> None:
        """@brief Handle cleaning up some of the connect modes.

        This init task de-asserts reset if the connect mode is under-reset.
        """
        mode = self.session.options.get('connect_mode')
        if mode == 'under-reset':
            LOG.info("Deasserting reset post connect")
            self.dp.assert_reset(False)

            LOG.debug("Clearing reset catch")
            # Apply to all cores.
            for core in self.cores.values():
                try:
                    core.clear_reset_catch(Target.ResetType.HW)
                except exceptions.Error as err:
                    LOG.warning("Could not halt core #%d: %s", core.core_number, err,
                        exc_info=self.session.log_tracebacks)

    def create_flash(self) -> None:
        """@brief Instantiates flash objects for memory regions.

        This init task iterates over flash memory regions and for each one finishes its setup and creates
        the Flash instance. It uses the flash_algo and flash_class properties of the region to know how
        to construct the flash object.
        """
        flm_builder = FlmFlashRegionBuilder(self, self.memory_map)
        for region in self.memory_map.iter_matching_regions(type=MemoryType.FLASH):
            region = cast(FlashRegion, region)

            # Load FLM file if needed, and create subregions for sector sizes.
            if not flm_builder.finalise_region(region):
                # Some error occurred, skip this region.
                continue

            # If the constructor of the region's flash class takes the flash_algo arg, then we
            # need the region to have a flash algo dict to pass to it. Otherwise we assume the
            # algo is built-in.
            klass = region.flash_class
            argspec = getfullargspec(klass.__init__)
            if 'flash_algo' in argspec.args:
                if region.algo is not None:
                    obj = klass(self, region.algo)
                else:
                    LOG.warning("flash region '%s' has no flash algo" % region.name)
                    continue
            else:
                obj = klass(self) # type:ignore

            # Set the region in the flash instance.
            obj.region = region

            # Store the flash object back into the memory region.
            region.flash = obj

        # Update the memory map in each core.
        for core in self.cores.values():
            core.memory_map = self.memory_map

    def check_for_cores(self) -> None:
        """@brief Init task: verify that at least one core was discovered."""
        if not len(self.cores):
            # Allow the user to override the exception to enable uses like chip bringup.
            if self.session.options.get('allow_no_cores'):
                LOG.error("No cores were discovered!")
            else:
                raise exceptions.DebugError("No cores were discovered!")

    @property
    def irq_table(self) -> Dict[int, str]:
        if (self._irq_table is None):
            if (self.svd_device is not None) and (self.svd_device.peripherals is not None):
                peripherals = [
                        p for p in self.svd_device.peripherals
                        if p.interrupts is not None
                        ]
                self._irq_table = {
                        i.value : i.name
                        for p in peripherals
                            for i in p.interrupts
                        }
            else:
                self._irq_table = {}
        return self._irq_table

    # Override this method from SoCTarget so we can use the DP for hardware resets when there isn't a
    # valid core (instead of the probe), so reset notifications will be sent. We can't use the DP in
    # SoCTarget because it is only created by this class.
    def reset(self, reset_type=None):
        # Use the DP to reset if there is not a core.
        if (self.selected_core is None) and (self.dp is not None):
            self.dp.reset()
        else:
            super().reset(reset_type)

    @property
    def first_ap(self) -> Optional["AccessPort"]:
        if len(self.aps) == 0:
            return None
        return sorted(self.aps.values(), key=lambda v: v.address)[0]

    def trace_start(self):
        result = self.call_delegate('trace_start', target=self, mode=0)
        if not result and self.has_debug_sequence('TraceStart', pname=self.selected_core_or_raise.node_name):
            assert self.debug_sequence_delegate
            self.debug_sequence_delegate.run_sequence('TraceStart',
                    pname=self.selected_core_or_raise.node_name)
            result = True
        return result

    def trace_stop(self):
        result = self.call_delegate('trace_stop', target=self, mode=0)
        if not result and self.has_debug_sequence('TraceStop', pname=self.selected_core_or_raise.node_name):
            assert self.debug_sequence_delegate
            self.debug_sequence_delegate.run_sequence('TraceStop',
                    pname=self.selected_core_or_raise.node_name)
            result = True
        return result




# pyOCD debugger
# Copyright (c) 2016-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

class Notification(object):
    """@brief Holds information about a notification to subscribers."""

    def __init__(self, event, source, data=None):
        self._event = event
        self._source = source
        self._data = data

    @property
    def event(self):
        return self._event

    @property
    def source(self):
        return self._source

    @property
    def data(self):
        return self._data

    def __repr__(self):
        return "<Notification@0x%08x event=%s source=%s data=%s>" % (id(self), repr(self.event), repr(self.source), repr(self.data))

class Notifier(object):
    """@brief Mix-in class that provides notification broadcast capabilities.

    In this notification model, subscribers register callbacks for one or more events. The events
    are simply a Python object of any kind, as long as it is hashable. Typically integers or Enums
    are used. Subscriptions can be registered for any sender of an event, or be filtered by the
    sender (called the source).

    When a notification is sent to the callback, it is wrapped up as a Notification object. Along
    with the notification, an optional, arbitrary data value can be sent. This allows for further
    specifying the event, or passing related values (or anything else you can think of).
    """

    def __init__(self):
        ## Dict of subscribers for particular events and sources.
        #
        # One subscriber can appear in multiple places in the dict.
        #
        # Schema:
        # ```
        # {
        #   event : ( [ subscribers-for-any-source ],
        #               {
        #                   source : [ subscribers ],
        #               } ),
        # }
        # ```
        self._subscribers = {}

    def subscribe(self, cb, events, source=None):
        """@brief Subscribe to selection of events from an optional source.

        @param self
        @param cb The callable that will be invoked when a matching notification is sent. Must
            accept a single parameter, a Notification instance.
        @param events Either a single event or an iterable of events. Events must be a hashable, and
            are usually just integers.
        @param source Optional notifier object. If not None, the callback is only invoked if one
            of the events is sent from the specified source. If a matching event is sent from
            another source, no action is taken.
        """
        if not isinstance(events, (tuple, list, set)):
            events = [events]

        for event in events:
            if event not in self._subscribers:
                self._subscribers[event] = ([], {})
            event_info = self._subscribers[event]

            if source is None:
                event_info[0].append(cb)
            else:
                if source not in event_info[1]:
                    event_info[1][source] = []
                event_info[1][source].append(cb)

    def unsubscribe(self, cb, events=None):
        """@brief Remove a callback from the subscribers list.

        @param self
        @param cb The callback to remove from all subscriptions.
        @param events Optional. May be a single event or an iterable of events. If specified, the
            _cb_ will be removed only from those events.
        """
        if (events is not None) and (not isinstance(events, (tuple, list, set))):
            events = [events]

        for event, event_info in self._subscribers.items():
            # Skip this event if it's not one on the removal list.
            if (events is not None) and (event not in events):
                continue

            # Remove callback from all-sources list.
            if cb in event_info[0]:
                event_info[0].remove(cb)

            # Scan source-specific subscribers.
            for source_info in event_info[1].values():
                if cb in source_info:
                    source_info.remove(cb)

    def notify(self, event, source=None, data=None):
        """@brief Notify subscribers of an event.

        @param self
        @param event Event to send. Must be a hashable object. It is acceptable to notify for an
            event for which there are no subscribers.
        @param source The object sending the notification. If not set, the source defaults to self,
            the object on which the notify() method was called.
        @param data Optional data value to send with the notification.
        """
        # Look up subscribers for this event.
        try:
            event_info = self._subscribers[event]
        except KeyError:
            # Nobody has subscribed to this event, so nothing to do.
            TRACE.debug("Not sending notification because no subscribers: event=%s", event)
            return

        # Look up subscribers for this event + source combo.
        try:
            source_subscribers = event_info[1][source]
        except KeyError:
            # No source-specific subscribers.
            source_subscribers = []

        # Create combined subscribers list. Exit if no subscribers matched.
        subscribers = event_info[0] + source_subscribers
        if not subscribers:
            TRACE.debug("Not sending notification because no matching subscribers: event=%s", event)
            return

        # Create the notification object now that we know there are some subscribers.
        if source is None:
            source = self
        note = Notification(event, source, data)
        TRACE.debug("Sending notification to %d subscribers: %s", len(subscribers), note)

        # Tell everyone!
        for cb in subscribers:
            cb(note)





# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import colorama
import io
import logging
import os
import traceback
from typing import (IO, Optional, Sequence, TYPE_CHECKING, Union)

from ..core.helpers import ConnectHelper
from ..core import (exceptions, session)
from ..probe.shared_probe_proxy import SharedDebugProbeProxy
from ..utility.cmdline import convert_session_options
from ..commands.repl import (PyocdRepl, ToolExitException)
from ..commands.execution_context import CommandExecutionContext

if TYPE_CHECKING:
    import argparse

LOG = logging.getLogger(__name__)

## Default SWD clock in Hz.
DEFAULT_CLOCK_FREQ_HZ = 1000000

class PyOCDCommander:
    """@brief Manages the commander interface.

    Responsible for connecting the execution context, REPL, and commands, and handles connection.

    Exit codes:
    - 0 = no errors
    - 1 = command error
    - 2 = transfer error
    - 3 = failed to create session (probe might not exist)
    - 4 = failed to open probe

    @todo Replace use of args from argparse with something cleaner.
    """

    CommandsListType = Sequence[Union[str, IO[str]]]

    ## Commands that can run without requiring a connection.
    _CONNECTIONLESS_COMMANDS = ('list', 'help', 'exit')

    def __init__(
                self,
                args: "argparse.Namespace",
                cmds: Optional[CommandsListType] = None
            ) -> None:
        """@brief Constructor."""
        # Read command-line arguments.
        self.args = args
        self.cmds: PyOCDCommander.CommandsListType = cmds or []

        self.context = CommandExecutionContext(no_init=self.args.no_init)
        self.context.command_set.add_command_group('commander')
        self.session: Optional[session.Session] = None
        self.exit_code: int = 0

    def run(self) -> int:
        """@brief Main entry point."""
        try:
            # If no commands, enter interactive mode. If there are commands, use the --interactive arg.
            enter_interactive = (not self.cmds) or self.args.interactive

            # Connect unless we are only running commands that don't require a connection.
            do_connect = enter_interactive or self._commands_require_connect()
            if do_connect and not self.connect():
                return self.exit_code

            # Run the list of commands we were given.
            if self.cmds:
                self.run_commands()

            # Enter the interactive REPL.
            if enter_interactive:
                assert self.session
                assert self.session.board
                assert self.context.target

                # Print connected message, unless not initing.
                if not self.args.no_init:
                    try:
                        # If the target is locked, we can't read the CPU state.
                        if self.session.target.is_locked():
                            status = "locked"
                        else:
                            try:
                                status = self.session.target.get_state().name.capitalize()
                            except (AttributeError, KeyError):
                                status = "<no core>"
                    except exceptions.TransferFaultError:
                        status = "<error>"
                else:
                    # Say what we're connected to, but without status.
                    status = "no init mode"

                # Say what we're connected to.
                print(colorama.Fore.GREEN + f"Connected to {self.session.target.part_number} " +
                        colorama.Fore.CYAN + f"[{status}]" +
                        colorama.Style.RESET_ALL + f": {self.session.board.unique_id}")

                # Run the REPL interface.
                console = PyocdRepl(self.context)
                console.run()

        except ToolExitException:
            self.exit_code = 0
        except exceptions.TransferError:
            print("Error: memory transfer failed")
            # Use get_current() in case our session hasn't been created yet.
            if session.Session.get_current().log_tracebacks:
                traceback.print_exc()
            self.exit_code = 2
        except exceptions.CommandError as e:
            print("Error:", e)
            self.exit_code = 1
        finally:
            # Ensure the session is closed.
            if self.session is not None:
                self.session.close()

        return self.exit_code

    def _commands_require_connect(self) -> bool:
        """@brief Determine whether a connection is needed to run commands."""
        for args in self.cmds:
            # Always assume connection required for command files.
            if isinstance(args, io.IOBase):
                return True

            # Check for connectionless commands.
            else:
                assert isinstance(args, str)

                if not ((len(args) == 1) and (args[0].lower() in self._CONNECTIONLESS_COMMANDS)):
                    return True

        # No command was found that needs a connection.
        return False

    def run_commands(self) -> None:
        """@brief Run commands specified on the command line."""
        for args in self.cmds:
            # Open file containing commands.
            if isinstance(args, io.IOBase) and not isinstance(args, str):
                self.context.process_command_file(args)

            # List of command and argument strings.
            else:
                assert isinstance(args, str)

                # Skip empty args lists.
                if len(args) == 0:
                    continue

                # Run the command line.
                self.context.process_command_line(args)

    def connect(self) -> bool:
        """@brief Connect to the probe."""
        if (self.args.frequency is not None) and (self.args.frequency != DEFAULT_CLOCK_FREQ_HZ):
            self.context.writei("Setting SWD clock to %d kHz", self.args.frequency // 1000)

        options = convert_session_options(self.args.options)

        # Set connect mode. The --connect option takes precedence when set. Then, if --halt is set
        # then the connect mode is halt. If connect_mode is set through -O then use that.
        # Otherwise default to attach.
        if hasattr(self.args, 'connect_mode') and self.args.connect_mode is not None:
            connect_mode = self.args.connect_mode
        elif self.args.halt:
            connect_mode = 'halt'
        elif 'connect_mode' in options:
            connect_mode = None
        else:
            connect_mode = 'attach'

        # Connect to board.
        probe = ConnectHelper.choose_probe(
                        blocking=(not self.args.no_wait),
                        unique_id=self.args.unique_id,
                        )
        if probe is None:
            self.exit_code = 3
            return False

        # Create a proxy so the probe can be shared between the session and a possible probe server.
        probe_proxy = SharedDebugProbeProxy(probe)

        # Create the session.
        self.session = session.Session(probe_proxy,
                        project_dir=self.args.project_dir,
                        config_file=self.args.config,
                        user_script=self.args.script,
                        no_config=self.args.no_config,
                        pack=self.args.pack,
                        target_override=self.args.target_override,
                        connect_mode=connect_mode,
                        frequency=self.args.frequency,
                        options=options,
                        option_defaults={
                            'auto_unlock': False,
                            'resume_on_disconnect': False,
                            'debug.traceback': logging.getLogger('pyocd').isEnabledFor(logging.DEBUG),
                            }
                        )

        if not self._post_connect():
            self.exit_code = 4
            return False

        result = self.context.attach_session(self.session)
        if not result:
            self.exit_code = 1
        return result

    def _post_connect(self) -> bool:
        """@brief Finish the connect process.

        The session is opened. The `no_init` parameter passed to the constructor determines whether the
        board and target are initialized.

        If an ELF file was provided on the command line, it is set on the target.

        @param self This object.
        @param session A @ref pyocd.core.session.Session "Session" instance.
        @retval True Session attached and context state inited successfully.
        @retval False An error occurred when opening the session or initing the context state.
        """
        assert self.session is not None
        assert not self.session.is_open

        # Open the session.
        try:
            self.session.open(init_board=not self.args.no_init)
        except exceptions.TransferFaultError as e:
            if not self.session.target.is_locked():
                LOG.error("Transfer fault while initing board: %s", e, exc_info=self.session.log_tracebacks)
                return False
        except exceptions.Error as e:
            LOG.error("Error while initing target: %s", e, exc_info=self.session.log_tracebacks)
            return False

        # Set elf file if provided.
        if self.args.elf:
            self.session.target.elf = os.path.expanduser(self.args.elf)

        # Handle a device with flash security enabled.
        if not self.args.no_init and self.session.target.is_locked():
            self.context.write("Warning: Target is locked, limited operations available. Use 'unlock' "
                                "command to mass erase and unlock, then execute 'reinit'.")

        return True




# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import prettytable
from typing import (cast, TYPE_CHECKING)

from .. import coresight
from ..core import exceptions
from ..probe.debug_probe import DebugProbe
from ..coresight.ap import MEM_AP
from ..core.target import Target
from ..core.core_target import CoreTarget
from ..utility.cmdline import (
    convert_one_session_option,
    convert_frequency,
    convert_vector_catch,
    convert_reset_type,
    )
from ..utility.mask import msb
from .base import ValueBase

if TYPE_CHECKING:
    from ..core.memory_map import MemoryMap
    from ..utility.graph import GraphNode

LOG = logging.getLogger(__name__)

VC_NAMES_MAP = {
        Target.VectorCatch.HARD_FAULT: "hard fault",
        Target.VectorCatch.BUS_FAULT: "bus fault",
        Target.VectorCatch.MEM_FAULT: "memory fault",
        Target.VectorCatch.INTERRUPT_ERR: "interrupt error",
        Target.VectorCatch.STATE_ERR: "state error",
        Target.VectorCatch.CHECK_ERR: "check error",
        Target.VectorCatch.COPROCESSOR_ERR: "coprocessor error",
        Target.VectorCatch.CORE_RESET: "core reset",
        Target.VectorCatch.SECURE_FAULT: "secure fault",
        }

HPROT_BIT_DESC = {
        0: ("instruction fetch", "data access"),
        1: ("user", "privileged"),
        2: ("non-bufferable", "bufferable"),
        3: ("non-cacheable", "cacheable/modifiable"),
        4: ("no cache lookup", "lookup in cache"),
        5: ("no cache allocate", "allocate in cache"),
        6: ("non-shareable", "shareable"),
        }

class UniqueIdValue(ValueBase):
    INFO = {
            'names': ['probe-uid', 'uid'],
            'group': 'standard',
            'category': 'probe',
            'access': 'r',
            'help': "Target's unique ID.",
            }

    def display(self, args):
        self.context.writei("Unique ID:    %s", self.context.board.unique_id)

class TargetValue(ValueBase):
    INFO = {
            'names': ['target'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "General target information.",
            }

    def display(self, args):
        self.context.write(f"Target type:  {self.context.session.options.get('target_override')}")
        self.context.write(f"Vendor:       {self.context.target.vendor}")
        self.context.write(f"Part number:  {self.context.target.part_number}")
        self.context.write(f"DAP IDCODE:   {self.context.target.dp.dpidr.idr:#010x}")

class CoresValue(ValueBase):
    INFO = {
            'names': ['cores'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Information about CPU cores in the target.",
            }

    def display(self, args):
        if self.context.target.is_locked():
            self.context.write("Target is locked")
        else:
            pt = prettytable.PrettyTable(["Number", "Name", "Type"])
            pt.align = 'l'
            pt.border = False

            for i, core in self.context.target.cores.items():
                pt.add_row([
                    (
                        ("*" if ((self.context.selected_core is not None)
                                and (self.context.selected_core.core_number == i))
                            else " ")
                        + str(i)
                    ),
                    cast(CoreTarget, core).node_name,
                    core.name,
                ])

            self.context.write(str(pt))
            self.context.write("(Currently selected core is marked with a '*'.)")

class APsValue(ValueBase):
    INFO = {
            'names': ['aps'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "List discovered Access Ports.",
            }

    def display(self, args):
        if self.context.target.is_locked():
            self.context.write("Target is locked")
        else:
            self.context.writei("%d APs:", len(self.context.target.aps))
            for addr, ap in sorted(self.context.target.aps.items(), key=lambda x: x[0]):
                self.context.writei("%s: %s%s", addr, ap.type_name,
                        " (selected)" if (self.context.selected_ap_address == addr) else "")

class MemoryMapValue(ValueBase):
    INFO = {
            'names': ['map'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Target memory map.",
            }

    def display(self, args):
        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return

        pt = prettytable.PrettyTable(["Region", "Type", "Start", "End", "Size", "Access", "Sector", "Page"])
        pt.align = 'l'
        pt.border = False

        def add_rows(indent: int, pt: prettytable.PrettyTable, map: "MemoryMap") -> None:
            for region in map:
                pt.add_row([
                    ('  ' * indent) + region.name,
                    region.type.name.capitalize(),
                    "0x%08x" % region.start,
                    "0x%08x" % region.end,
                    "0x%08x" % region.length,
                    region.access,
                    ("0x%08x" % region.sector_size) if region.is_flash else '-',
                    ("0x%08x" % region.page_size) if region.is_flash else '-',
                    ])
                if region.has_subregions:
                    add_rows(indent + 2, pt, region.submap)

        add_rows(0, pt, self.context.selected_core.memory_map)

        self.context.write(str(pt))

class PeripheralsValue(ValueBase):
    INFO = {
            'names': ['peripherals'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "List of target peripheral instances.",
            }

    def display(self, args):
        for periph in sorted(self.context.peripherals.values(), key=lambda x:x.base_address):
            self.context.writei("0x%08x: %s", periph.base_address, periph.name)

class FaultValue(ValueBase):
    INFO = {
            'names': ['fault'],
            'group': 'standard',
            'category': 'exceptions',
            'access': 'r',
            'show_usage': "[-a]",
            'help': "Fault status information.",
            'extra_help': "By default, only asserted fields are shown. Add -a to command to show all fields.",
            }

    def display(self, args):
        showAll = ('-a' in args)

        CFSR = 0xe000ed28
        HFSR = 0xe000ed2c
        DFSR = 0xe000ed30
        MMFAR = 0xe000ed34
        BFAR = 0xe000ed38
#         AFSR = 0xe000ed3c

        MMFSR_fields = [
                ('IACCVIOL', 0),
                ('DACCVIOL', 1),
                ('MUNSTKERR', 3),
                ('MSTKERR', 4),
#                 ('MMARVALID', 7),
            ]
        BFSR_fields = [
                ('IBUSERR', 0),
                ('PRECISERR', 1),
                ('IMPRECISERR', 2),
                ('UNSTKERR', 3),
                ('STKERR', 4),
                ('LSPERR', 5),
#                 ('BFARVALID', 7),
            ]
        UFSR_fields = [
                ('UNDEFINSTR', 0),
                ('INVSTATE', 1),
                ('INVPC', 2),
                ('NOCP', 3),
                ('STKOF', 4),
                ('UNALIGNED', 8),
                ('DIVBYZERO', 9),
            ]
        HFSR_fields = [
                ('VECTTBL', 1),
                ('FORCED', 30),
                ('DEBUGEVT', 31),
            ]
        DFSR_fields = [
                ('HALTED', 0),
                ('BKPT', 1),
                ('DWTTRAP', 2),
                ('VCATCH', 3),
                ('EXTERNAL', 4),
            ]

        def print_fields(regname, value, fields, showAll):
            if value == 0 and not showAll:
                return
            self.context.writei("  %s = 0x%08x", regname, value)
            for name, bitpos in fields:
                bit = (value >> bitpos) & 1
                if showAll or bit != 0:
                    self.context.writei("    %s = 0x%x", name, bit)

        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return

        cfsr = self.context.selected_core.read32(CFSR)
        mmfsr = cfsr & 0xff
        bfsr = (cfsr >> 8) & 0xff
        ufsr = (cfsr >> 16) & 0xffff
        hfsr = self.context.selected_core.read32(HFSR)
        dfsr = self.context.selected_core.read32(DFSR)
        mmfar = self.context.selected_core.read32(MMFAR)
        bfar = self.context.selected_core.read32(BFAR)

        print_fields('MMFSR', mmfsr, MMFSR_fields, showAll)
        if showAll or mmfsr & (1 << 7): # MMFARVALID
            self.context.writei("  MMFAR = 0x%08x", mmfar)
        print_fields('BFSR', bfsr, BFSR_fields, showAll)
        if showAll or bfsr & (1 << 7): # BFARVALID
            self.context.writei("  BFAR = 0x%08x", bfar)
        print_fields('UFSR', ufsr, UFSR_fields, showAll)
        print_fields('HFSR', hfsr, HFSR_fields, showAll)
        print_fields('DFSR', dfsr, DFSR_fields, showAll)

class NresetValue(ValueBase):
    INFO = {
            'names': ['nreset'],
            'group': 'standard',
            'category': 'target',
            'access': 'rw',
            'help': "Current nRESET signal state.",
            'extra_help': "Accepts a value of 0 or 1.",
            }

    def display(self, args):
        rst = int(not self.context.probe.is_reset_asserted())
        self.context.writef("nRESET = {}", rst)

    def modify(self, args):
        if len(args) != 1:
            raise exceptions.CommandError("missing reset state")
        state = int(args[0], base=0)
        self.context.writef("nRESET = {}", state)

        # Use the probe to assert reset if the DP doesn't exist for some reason, otherwise
        # use the DP so reset notifications are sent.
        if self.context.target.dp is None:
            self.context.probe.assert_reset((state == 0))
        else:
            self.context.target.dp.assert_reset((state == 0))

class AccessiblePinsValue(ValueBase):
    INFO = {
            'names': ['accessible-pins'],
            'group': 'standard',
            'category': 'probe',
            'access': 'rw',
            'help': "Display which debug probe pins can be read and written with the 'pins' value.",
            }

    def display(self, args):
        if DebugProbe.Capability.PIN_ACCESS not in self.context.probe.capabilities:
            raise exceptions.CommandError("debug probe does not support pin access")

        # Display accessibility of protocol pins.
        r_pins, w_pins = self.context.probe.get_accessible_pins(DebugProbe.PinGroup.PROTOCOL_PINS)

        def pin_desc(mask: int) -> str:
            desc = ""
            if r_pins & mask:
                desc += "r"
            if w_pins & mask:
                desc += "w"
            if desc == "":
                desc = "n/a"
            return desc

        self.context.write(f"Protocol pins:")
        self.context.write(f"  SWCLK/TCK = {pin_desc(DebugProbe.ProtocolPin.SWCLK_TCK)}")
        self.context.write(f"  SWDIO/TMS = {pin_desc(DebugProbe.ProtocolPin.SWDIO_TMS)}")
        self.context.write(f"  TDI =       {pin_desc(DebugProbe.ProtocolPin.TDI)}")
        self.context.write(f"  TDO =       {pin_desc(DebugProbe.ProtocolPin.TDO)}")
        self.context.write(f"  nRESET =    {pin_desc(DebugProbe.ProtocolPin.nRESET)}")
        self.context.write(f"  nTRST =     {pin_desc(DebugProbe.ProtocolPin.nTRST)}")

        # Test if there are any accessible GPIO pins.
        r_pins, w_pins = self.context.probe.get_accessible_pins(DebugProbe.PinGroup.GPIO_PINS)

        if (r_pins | w_pins) != 0:
            self.context.write(f"GPIO pins:")
            for b in range(msb(r_pins | w_pins) + 1):
                self.context.write(f"  GPIO {b:<6} {pin_desc(1 << b)}")

class PinsValue(ValueBase):
    INFO = {
            'names': ['pins'],
            'group': 'standard',
            'category': 'probe',
            'access': 'rw',
            'help': "Current debug probe protocol I/O pin states.",
            'extra_help':
                "The pins value is a mask containing the state of all accessible protocol pins. "
                "See the `accessible-pins` value for protocol pins that can be read and written by "
                "the connected debug probe.",
            }

    def display(self, args):
        if DebugProbe.Capability.PIN_ACCESS not in self.context.probe.capabilities:
            raise exceptions.CommandError("debug probe does not support pin access")
        self.print_current_pin_values()

    def print_current_pin_values(self):
        pins = self.context.probe.read_pins(DebugProbe.PinGroup.PROTOCOL_PINS,
                                            DebugProbe.ProtocolPin.ALL_PINS)

        def pin_desc(mask: int) -> str:
            v = int(pins & mask != 0)
            return f"{v} (mask {mask:#x})"

        self.context.write(f"Pins mask = {pins:#x}")
        self.context.write(f"SWCLK/TCK = {pin_desc(DebugProbe.ProtocolPin.SWCLK_TCK)}")
        self.context.write(f"SWDIO/TMS = {pin_desc(DebugProbe.ProtocolPin.SWDIO_TMS)}")
        self.context.write(f"TDI =       {pin_desc(DebugProbe.ProtocolPin.TDI)}")
        self.context.write(f"TDO =       {pin_desc(DebugProbe.ProtocolPin.TDO)}")
        self.context.write(f"nRESET =    {pin_desc(DebugProbe.ProtocolPin.nRESET)}")
        self.context.write(f"nTRST =     {pin_desc(DebugProbe.ProtocolPin.nTRST)}")

    def modify(self, args):
        if DebugProbe.Capability.PIN_ACCESS not in self.context.probe.capabilities:
            raise exceptions.CommandError("debug probe does not support pin access")
        if len(args) != 1:
            raise exceptions.CommandError("missing pins mask")
        state = int(args[0], base=0)
        self.context.probe.write_pins(DebugProbe.PinGroup.PROTOCOL_PINS,
                                        DebugProbe.ProtocolPin.ALL_PINS, state)
        self.print_current_pin_values()

class SessionOptionValue(ValueBase):
    INFO = {
            'names': ['option'],
            'group': 'standard',
            'category': 'options',
            'access': 'rw',
            'show_usage': "NAME+",
            'set_usage': "NAME[=VALUE]+",
            'help': "The current value of one or more session options.",
            'extra_help': "When setting, each argument should follow the form \"NAME[=VALUE]\".",
            }

    def display(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("missing session option name argument")
        for name in args:
            try:
                value = self.context.session.options[name]
                self.context.writei("Option '%s' = %s", name, value)
            except KeyError:
                self.context.writei("No option with name '%s'", name)

    def modify(self, args):
        """Extract and apply option setting arguments.

        The syntax for each option is "name[=value]". The args are pre-split into individual tokens,
        where the '=' is a separate token. So a single "foo=bar" is split into "foo", "=", "bar" args.
        """
        if len(args) < 1:
            raise exceptions.CommandError("missing session option setting")
        while args:
            name = args.pop(0)
            if args and args[0] == "=":
                args.pop(0) # Remove "="
                if not args:
                    raise exceptions.CommandError("expected option value after '='")
                value = args.pop(0)
            else:
                value = None
            name, converted_value = convert_one_session_option(name, value)
            self.context.session.options[name] = converted_value

class MemApValue(ValueBase):
    INFO = {
            'names': ['mem-ap'],
            'group': 'standard',
            'category': 'memory',
            'access': 'rw',
            'help': "The currently selected MEM-AP used for memory read/write commands.",
            'extra_help': "When the selected core is changed by the 'core' command, the selected "
                "MEM-AP is changed to match. This overrides a user-selected MEM-AP if different "
                "from the AP for the newly selected core.",
            }

    def display(self, args):
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        self.context.writef("{} is selected", self.context.selected_ap.short_description)

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")

        ap_num = int(args[0], base=0)
        if self.context.target.dp.adi_version == coresight.dap.ADIVersion.ADIv5:
            ap_addr = coresight.ap.APv1Address(ap_num)
        elif self.context.target.dp.adi_version == coresight.dap.ADIVersion.ADIv6:
            ap_addr = coresight.ap.APv2Address(ap_num)
        if ap_addr not in self.context.target.aps:
            self.context.writef("Invalid AP number {:#x}", ap_num)
            return
        ap = self.context.target.aps[ap_addr]
        if not isinstance(ap, MEM_AP):
            self.context.writef("{} is not a MEM-AP", ap.short_description)
            return
        self.context.selected_ap_address = ap_addr

class HnonsecValue(ValueBase):
    INFO = {
            'names': ['hnonsec'],
            'group': 'standard',
            'category': 'memory',
            'access': 'rw',
            'help': "The current HNONSEC attribute value used by the selected MEM-AP.",
            'extra_help':
                "This value controls whether memory transactions are secure or nonsecure. The value is an "
                "integer, either 0 or secure or 1 for nonsecure."
            }

    def display(self, args):
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        self.context.writef("{} HNONSEC = {} ({})",
            self.context.selected_ap.short_description,
            self.context.selected_ap.hnonsec,
            ("nonsecure" if self.context.selected_ap.hnonsec else "secure"))

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        value = int(args[0], base=0)
        self.context.selected_ap.hnonsec = value

class HprotValue(ValueBase):
    INFO = {
            'names': ['hprot', 'memap_attr'],
            'group': 'standard',
            'category': 'memory',
            'access': 'rw',
            'help': "The current memory transfer attributes value used by the selected MEM-AP.",
            'extra_help':
"""This integer value controls attributes of memory transfers. It is a direct mapping of the AHB
or AXI attribute settings, depending on the type of MEM-AP. For AHB-APs, the value is HPROT[4:0].
For AXI-APs, the value is {AxPROT[2:0}, AxCACHE[3:0]}, e.g. AxPROT in bits 6-4 and AxCACHE in
its 3-0. Not all MEM-AP implementations support all attributes. See the Arm Technical Reference
Manual for your device's MEM-AP for details."""
            }

    def display(self, args):
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        hprot = self.context.selected_ap.hprot
        self.context.writef("{} HPROT = {:#x}",
            self.context.selected_ap.short_description,
            hprot)
        desc = ""
        for bitnum in range(7):
            bitvalue = (hprot >> bitnum) & 1
            desc += "    HPROT[{}] = {:#x} ({})\n".format(
                bitnum,
                bitvalue,
                HPROT_BIT_DESC[bitnum][bitvalue])
        self.context.write(desc, end='')

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")
        if self.context.selected_ap is None:
            self.context.write("No MEM-AP is selected")
            return
        value = int(args[0], base=0)
        self.context.selected_ap.hprot = value

class TargetGraphValue(ValueBase):
    INFO = {
            'names': ['graph'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Print the target object graph.",
            }

    def display(self, args):
        def _node_desc(node: GraphNode) -> str:
            desc = node.__class__.__name__
            if node.node_name:
                desc = f'"{node.node_name}": ' + desc
            return desc

        def _dump(node: GraphNode, level: int) -> None:
            desc = ("  " * level) + "- " + _node_desc(node)
            self.context.write(desc)
            for child in node.children:
                _dump(child, level + 1)

        _dump(self.context.board, 0)

class LockedValue(ValueBase):
    INFO = {
            'names': ['locked'],
            'group': 'standard',
            'category': 'target',
            'access': 'r',
            'help': "Report whether the target is locked.",
            }

    def display(self, args):
        if self.context.target.is_locked():
            self.context.write("Target is locked")
        else:
            self.context.write("Target is unlocked")

class RegisterGroupsValue(ValueBase):
    INFO = {
            'names': ['register-groups'],
            'group': 'standard',
            'category': 'registers',
            'access': 'r',
            'help': "Display available register groups for the selected core.",
            }

    def display(self, args):
        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return
        for g in sorted(self.context.selected_core.core_registers.groups):
            self.context.write(g)

class VectorCatchValue(ValueBase):
    INFO = {
            'names': ['vector-catch', 'vc'],
            'group': 'standard',
            'category': 'exceptions',
            'access': 'rw',
            'help': "Show current vector catch settings.",
            'extra_help': "When setting, the alue is a concatenation of one letter per enabled source in "
                "any order, or 'all' or 'none'. (h=hard fault, b=bus fault, m=mem fault, i=irq err, s=state "
                "err, c=check err, p=nocp, r=reset, a=all, n=none).",
            }

    def display(self, args):
        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return

        catch = self.context.selected_core.get_vector_catch()

        self.context.write("Vector catch:")
        for mask in sorted(VC_NAMES_MAP.keys()):
            name = VC_NAMES_MAP[mask]
            s = "ON" if (catch & mask) else "OFF"
            self.context.writef("  {:3} {}", s, name)

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing vector catch setting")

        if self.context.selected_core is None:
            self.context.write("No core is selected")
            return

        try:
            self.context.selected_core.set_vector_catch(convert_vector_catch(args[0]))
        except ValueError as e:
            self.context.write(e)

class StepIntoInterruptsValue(ValueBase):
    INFO = {
            'names': ['step-into-interrupts', 'si'],
            'group': 'standard',
            'category': 'options',
            'access': 'rw',
            'help': "Display whether interrupts are enabled when single stepping.",
            'extra_help': "Set to 1 to enable.",
            }

    def display(self, args):
        self.context.writei("Interrupts while stepping: %s",
            ("enabled" if self.context.session.options['step_into_interrupt'] else "disabled"))

    def modify(self, args):
        if len(args) == 0:
            raise exceptions.CommandError("missing argument")
        value = (args[0] in ('1', 'true', 'yes', 'on'))
        self.context.session.options['step_into_interrupt'] = value

class LogLevelValue(ValueBase):
    INFO = {
            'names': ['log'],
            'group': 'standard',
            'category': 'miscellaneous',
            'access': 'w',
            'set_usage': "LEVEL [MODULE+]",
            'help': "Set log level to one of 'debug', 'info', 'warning', 'error', 'critical'.",
            'extra_help': "If pyocd module names are provided as arguments after the log level then "
                            "only those modules will have their log level changed."
            }

    LEVELS = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
            }

    def display(self, args):
        pass

    def modify(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("no log level provided")
        level_name = args[0].lower()
        if level_name not in self.LEVELS:
            raise exceptions.CommandError("log level must be one of {%s}" % ','.join(self.LEVELS.keys()))
        level = self.LEVELS[level_name]
        if len(args) == 1:
            logging.getLogger().setLevel(level)
        else:
            for module_name in args[1:]:
                # Check if there is already a logger with this name.
                #
                # This is guarded with a check that the root logger has a 'manager' attribute, and the
                # the Manager object has a 'loggerDict' attribute. Both attributes are undocumented, so
                # a check is necessary to prevent failure in case the logging code changes.
                if hasattr(logging.root, 'manager'):
                    manager = logging.root.manager
                    if hasattr(manager, 'loggerDict') and (module_name not in manager.loggerDict):
                        self.context.writef("No logger exists with the name {}", module_name)
                        continue
                logging.getLogger(module_name).setLevel(level)

class ClockFrequencyValue(ValueBase):
    INFO = {
            'names': ['frequency'],
            'group': 'standard',
            'category': 'options',
            'access': 'w',
            'help': "Set SWD or JTAG clock frequency in Hertz.",
            'extra_help': "A case-insensitive metric scale suffix of either 'k' or 'm' is allowed, as well "
                "as a trailing \"Hz\". There must be no space between the frequency and the suffix. For "
                "example, \"2.5MHz\" sets the clock to 2.5 MHz.",
            }

    def display(self, args):
        pass

    def modify(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("no clock frequency provided")
        try:
            freq_Hz = convert_frequency(args[0])
        except Exception:
            raise exceptions.CommandError("invalid frequency")
        self.context.probe.set_clock(freq_Hz)
        if self.context.probe.wire_protocol == DebugProbe.Protocol.SWD:
            swd_jtag = 'SWD'
        elif self.context.probe.wire_protocol == DebugProbe.Protocol.JTAG:
            swd_jtag = 'JTAG'
        else:
            swd_jtag = '??'

        if freq_Hz >= 1000000:
            nice_freq = "%.2f MHz" % (freq_Hz / 1000000)
        elif freq_Hz > 1000:
            nice_freq = "%.2f kHz" % (freq_Hz / 1000)
        else:
            nice_freq = "%d Hz" % freq_Hz

        self.context.writei("Changed %s frequency to %s", swd_jtag, nice_freq)

class DebugSequencesValue(ValueBase):
    INFO = {
            'names': ['debug-sequences'],
            'group': 'pack-target',
            'category': 'cmsis-pack',
            'access': 'r',
            'help': "Show the available debug sequences from the target's DFP.",
            'extra_help': "Only available for CMSIS-Pack based targets.",
            }

    # Names of debug sequences that are both standard and supported by pyocd.
    STANDARD_SEQUENCE_NAMES = [
        "DebugPortSetup",
        "DebugPortStart",
        "DebugPortStop",
        "DebugDeviceUnlock",
        "DebugCoreStart",
        "DebugCoreStop",
        "ResetSystem",
        "ResetProcessor",
        "ResetHardware",
        "ResetCatchSet",
        "ResetCatchClear",
        "TraceStart",
        "TraceStop",
    ]

    def display(self, args):
        assert self.context.target

        if self.context.target.debug_sequence_delegate is None:
            self.context.write("Target does not use debug sequences")
            return

        pt = prettytable.PrettyTable(["Name", "Processor", "Standard", "Enabled"])
        pt.align = 'l'
        pt.border = False

        for seq in sorted(self.context.target.debug_sequence_delegate.all_sequences,
                key=lambda i: (i.name, i.pname)):
            is_standard = seq.name in self.STANDARD_SEQUENCE_NAMES
            pt.add_row([
                seq.name,
                seq.pname if seq.pname else "all",
                str(is_standard),
                seq.is_enabled,
            ])

        self.context.write(str(pt))

class ResetTypeValue(ValueBase):
    INFO = {
            'names': ['reset-type'],
            'group': 'standard',
            'category': 'target',
            'access': 'rw',
            'help': "Show reset configuration and all available reset types for each core. Set current reset type.",
            }

    def display(self, args):
        from ..coresight.cortex_m import CortexM

        assert self.context.target

        current_reset_type_option = self.context.session.options.get('reset_type')
        current_reset_type = convert_reset_type(current_reset_type_option)
        reset_type_desc = current_reset_type_option
        if current_reset_type is not None:
            reset_type_desc += f" ({current_reset_type.name})"
        self.context.write(f"Selected reset type ('reset_type' option): {reset_type_desc}")

        for core in self.context.target.cores.values():
            # Only handle Cortex-M cores for now.
            if not isinstance(core, CortexM):
                continue
            cm_core = cast(CortexM, core)

            actual_reset_type = cm_core._get_actual_reset_type(None)

            self.context.write(f"\nCore {cm_core.core_number} ({cm_core.node_name}):")
            self.context.write(f"  Effective:  {actual_reset_type.name}")
            self.context.write(f"  Default:    {cm_core.default_reset_type.name}")
            self.context.write(f"  Default SW: {cm_core.default_software_reset_type.name}")
            self.context.write("  Available:  " + ", ".join(r.name for r in cm_core.supported_reset_types))

    def modify(self, args):
        from ..utility.cmdline import RESET_TYPE_MAP
        if len(args) != 1:
            raise exceptions.CommandError("invalid arguments")

        new_reset_type = args[0]
        if new_reset_type.lower() not in RESET_TYPE_MAP:
            raise exceptions.CommandError("invalid reset type")

        self.context.session.options['reset_type'] = new_reset_type



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List
import logging
import json
import traceback

from .base import SubcommandBase
from ..core.session import Session
from ..tools.lists import ListGenerator
from ..target.pack import pack_target
from ..utility.cmdline import convert_session_options
from .. import __version__

LOG = logging.getLogger(__name__)

class JsonSubcommand(SubcommandBase):
    """@brief `pyocd json` subcommand."""

    NAMES = ['json']
    HELP = "Output information as JSON."
    DEFAULT_LOG_LEVEL = logging.FATAL + 1

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""
        json_parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        json_options = json_parser.add_argument_group('json output')
        json_options.add_argument('-p', '--probes', action='store_true',
            help="List available probes.")
        json_options.add_argument('-t', '--targets', action='store_true',
            help="List all known targets.")
        json_options.add_argument('-b', '--boards', action='store_true',
            help="List all known boards.")
        json_options.add_argument('-f', '--features', action='store_true',
            help="List available features and options.")

        return [cls.CommonOptions.CONFIG, json_parser]

    @classmethod
    def customize_subparser(cls, subparser: argparse.ArgumentParser) -> None:
        """@brief Optionally modify a subparser after it is created."""
        subparser.set_defaults(verbose=0, quiet=0)

    def __init__(self, args: argparse.Namespace):
        super().__init__(args)

        # Disable all logging.
        logging.disable(logging.CRITICAL)

    def invoke(self) -> int:
        """@brief Handle 'json' subcommand."""
        exit_status = 0
        try:
            all_outputs = (self._args.probes, self._args.targets, self._args.boards, self._args.features)

            # Default to listing probes.
            if not any(all_outputs):
                self._args.probes = True

            # Check for more than one output option being selected.
            if sum(int(x) for x in all_outputs) > 1:
                # Because we're outputting JSON we can't just log the error, but must report the error
                # via the JSON format.
                obj = {
                    'pyocd_version' : __version__,
                    'version' : { 'major' : 1, 'minor' : 0 },
                    'status' : 1,
                    'error' : "More than one output data selected.",
                    }
                exit_status = 1
            else:
                # Create a session with no device so we load any config.
                session = Session(None,
                                    project_dir=self._args.project_dir,
                                    config_file=self._args.config,
                                    no_config=self._args.no_config,
                                    pack=self._args.pack,
                                    **convert_session_options(self._args.options)
                                    )

                if self._args.targets or self._args.boards:
                    # Create targets from provided CMSIS pack.
                    if session.options['pack'] is not None:
                        pack_target.PackTargets.populate_targets_from_pack(session.options['pack'])

                if self._args.probes:
                    obj = ListGenerator.list_probes()
                elif self._args.targets:
                    obj = ListGenerator.list_targets()
                elif self._args.boards:
                    obj = ListGenerator.list_boards()
                elif self._args.features:
                    obj = ListGenerator.list_features()
                else:
                    assert False
        except Exception as e:
            # Report exceptions via JSON output.
            obj = {
                'pyocd_version' : __version__,
                'version' : { 'major' : 1, 'minor' : 0 },
                'status' : 1,
                'error' : f"Error occurred during processing.\n" + traceback.format_exc(),
                }
            exit_status = 1

        print(json.dumps(obj, indent=4))
        return exit_status




# pyOCD debugger
# Copyright (c) 2016-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class CacheMetrics(object):
    """@brief Holds hit ratio metrics for the caches."""
    def __init__(self):
        self.hits = 0
        self.misses = 0
        self.reads = 0
        self.writes = 0

    @property
    def total(self):
        return self.hits + self.misses

    @property
    def percent_hit(self):
        if self.total > 0:
            return self.hits * 100.0 / self.total
        else:
            return 0

    @property
    def percent_miss(self):
        if self.total > 0:
            return self.misses * 100.0 / self.total
        else:
            return 0




# pyOCD debugger
# Copyright (c) 2017 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..symbols import SymbolProvider

class ELFSymbolProvider(SymbolProvider):
    """@brief Get symbol information from an ELF file."""

    def __init__(self, elf):
        self._symbols = elf.symbol_decoder

    def get_symbol_value(self, name):
        sym = self._symbols.get_symbol_for_name(name)
        if sym is not None:
            return sym.address
        else:
            return None




# pyOCD debugger
# Copyright (c) 2016-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .provider import (TargetThread, ThreadProvider)
from .common import (read_c_string, HandlerModeThread, EXC_RETURN_EXT_FRAME_MASK)
from ..core import exceptions
from ..core.target import Target
from ..core.plugin import Plugin
from ..debug.context import DebugContext
from ..coresight.cortex_m_core_registers import index_for_reg
import logging

# Create a logger for this module.
LOG = logging.getLogger(__name__)

class TargetList(object):
    def __init__(self, context, ptr, nextOffset):
        self._context = context
        self._list = ptr
        self._offset = nextOffset

    def __iter__(self):
        # Read first item on list.
        node = self._context.read32(self._list)

        while node != 0:
            # Return previously read item.
            yield node

            try:
                # Read the next item in the list.
                node = self._context.read32(node + self._offset)
            except exceptions.TransferError as exc:
                LOG.warning("TransferError while reading list elements (list=0x%08x, node=0x%08x), terminating list: %s", self._list, node, exc)
                break

class RTXThreadContext(DebugContext):
    """@brief Thread context for RTX5."""

    # SP/PSP are handled specially, so it is not in these dicts.

    # Offsets are relative to stored SP in a task switch block, for the
    # combined software + hardware stacked registers. In exception case,
    # software registers are not stacked, so appropriate amount must be
    # subtracted.
    NOFPU_REGISTER_OFFSETS = {
                 # Software stacked
                 4: 0, # r4
                 5: 4, # r5
                 6: 8, # r6
                 7: 12, # r7
                 8: 16, # r8
                 9: 20, # r9
                 10: 24, # r10
                 11: 28, # r11
                 # Hardware stacked
                 0: 32, # r0
                 1: 36, # r1
                 2: 40, # r2
                 3: 44, # r3
                 12: 48, # r12
                 14: 52, # lr
                 15: 56, # pc
                 16: 60, # xpsr
            }

    FPU_REGISTER_OFFSETS = {
                 # Software stacked
                 0x50: 0, # s16
                 0x51: 4, # s17
                 0x52: 8, # s18
                 0x53: 12, # s19
                 0x54: 16, # s20
                 0x55: 20, # s21
                 0x56: 24, # s22
                 0x57: 28, # s23
                 0x58: 32, # s24
                 0x59: 36, # s25
                 0x5a: 40, # s26
                 0x5b: 44, # s27
                 0x5c: 48, # s28
                 0x5d: 52, # s29
                 0x5e: 56, # s30
                 0x5f: 60, # s31
                 4: 64, # r4
                 5: 68, # r5
                 6: 72, # r6
                 7: 76, # r7
                 8: 80, # r8
                 9: 84, # r9
                 10: 88, # r10
                 11: 92, # r11
                 # Hardware stacked
                 0: 96, # r0
                 1: 100, # r1
                 2: 104, # r2
                 3: 108, # r3
                 12: 112, # r12
                 14: 116, # lr
                 15: 120, # pc
                 16: 124, # xpsr
                 0x40: 128, # s0
                 0x41: 132, # s1
                 0x42: 136, # s2
                 0x43: 140, # s3
                 0x44: 144, # s4
                 0x45: 148, # s5
                 0x46: 152, # s6
                 0x47: 156, # s7
                 0x48: 160, # s8
                 0x49: 164, # s9
                 0x4a: 168, # s10
                 0x4b: 172, # s11
                 0x4c: 176, # s12
                 0x4d: 180, # s13
                 0x4e: 184, # s14
                 0x4f: 188, # s15
                 33: 192, # fpscr
                 # (reserved word: 196)
            }

    def __init__(self, parent, thread):
        super(RTXThreadContext, self).__init__(parent)
        self._thread = thread
        self._has_fpu = self.core.has_fpu

    def read_core_registers_raw(self, reg_list):
        reg_list = [index_for_reg(reg) for reg in reg_list]
        reg_vals = []

        isCurrent = self._thread.is_current
        inException = isCurrent and self._parent.read_core_register('ipsr') > 0

        # If this is the current thread and we're not in an exception, just read the live registers.
        if isCurrent and not inException:
            return self._parent.read_core_registers_raw(reg_list)

        # Because of above tests, from now on, inException implies isCurrent;
        # we are generating the thread view for the RTOS thread where the
        # exception occurred; the actual Handler Mode thread view is produced
        # by HandlerModeThread
        if inException:
            # Reasonable to assume PSP is still valid
            sp = self._parent.read_core_register('psp')
        else:
            sp = self._thread.get_stack_pointer()

        # Determine which register offset table to use and the offsets past the saved state.
        hwStacked = 0x20
        swStacked = 0x20
        table = self.NOFPU_REGISTER_OFFSETS
        if self._has_fpu:
            try:
                if inException and self.core.is_vector_catch():
                    # Vector catch has just occurred, take live LR
                    exceptionLR = self._parent.read_core_register('lr')
                else:
                    # Can't really rely on finding live LR after initial
                    # vector catch, so retrieve LR stored by OS on last
                    # thread switch.
                    exceptionLR = self._thread.get_stack_frame()

                # Check bit 4 of the exception LR to determine if FPU registers were stacked.
                if (exceptionLR & EXC_RETURN_EXT_FRAME_MASK) == 0:
                    table = self.FPU_REGISTER_OFFSETS
                    hwStacked = 0x68
                    swStacked = 0x60
            except exceptions.TransferError:
                LOG.debug("Transfer error while reading thread's saved LR")

        for reg in reg_list:

            # Must handle stack pointer specially.
            if reg == 13:
                if inException:
                    reg_vals.append(sp + hwStacked)
                else:
                    reg_vals.append(sp + swStacked + hwStacked)
                continue

            # Look up offset for this register on the stack.
            spOffset = table.get(reg, None)
            if spOffset is None:
                reg_vals.append(self._parent.read_core_register_raw(reg))
                continue
            if inException:
                spOffset -= swStacked

            try:
                if spOffset >= 0:
                    reg_vals.append(self._parent.read32(sp + spOffset))
                else:
                    # Not available - try live one
                    reg_vals.append(self._parent.read_core_register_raw(reg))
            except exceptions.TransferError:
                reg_vals.append(0)

        return reg_vals

class RTXTargetThread(TargetThread):
    """@brief Represents an RTX5 thread on the target."""

    STATE_OFFSET = 1
    NAME_OFFSET = 4
    PRIORITY_OFFSET = 33
    STACKFRAME_OFFSET = 34
    SP_OFFSET = 56

    STATES = {
         0x00: "Inactive",
         0x01: "Ready",
         0x02: "Running",
         0x03: "Blocked",
         0x04: "Terminated",
         0x13: "Waiting[Delay]",
         0x23: "Waiting[Join]",
         0x33: "Waiting[ThrFlg]",
         0x43: "Waiting[EvtFlg]",
         0x53: "Waiting[Mutex]",
         0x63: "Waiting[Sem]",
         0x73: "Waiting[MemPool]",
         0x83: "Waiting[MsgGet]",
         0x93: "Waiting[MsgPut]",
    }

    def __init__(self, targetContext, provider, base):
        super(RTXTargetThread, self).__init__()
        self._target_context = targetContext
        self._provider = provider
        self._base = base
        self._state = 0
        self._priority = 0
        self._thread_context = RTXThreadContext(self._target_context, self)
        self._has_fpu = self._thread_context.core.has_fpu
        try:
            name_ptr = self._target_context.read32(self._base + RTXTargetThread.NAME_OFFSET)
            self._name = read_c_string(self._target_context, name_ptr)

            self.update_state()
        except exceptions.TransferError as exc:
            LOG.debug("Transfer error while reading thread %x name: %s", self._base, exc)
            self._name = "?"
        LOG.debug('RTXTargetThread 0x%x' % base)

    def update_state(self):
        try:
            state = self._target_context.read8(self._base + RTXTargetThread.STATE_OFFSET)
            priority = self._target_context.read8(self._base + RTXTargetThread.PRIORITY_OFFSET)
        except exceptions.TransferError as exc:
            LOG.debug("Transfer error while reading thread %x state: %s", self._base, exc)
        else:
            self._state = state
            self._priority = priority

    @property
    def priority(self):
        return self._priority

    @property
    def unique_id(self):
        # There is no other meaningful ID than base address
        return self._base

    @property
    def context(self):
        return self._thread_context

    @property
    def description(self):
        return "%s; Priority %d" % (self.STATES.get(self._state, "(Invalid)"), self.priority)

    @property
    def name(self):
        return self._name

    @property
    def is_current(self):
        return self._provider.get_actual_current_thread_id() == self._base

    def get_stack_pointer(self):
        # Get stack pointer saved in thread struct.
        try:
            return self._target_context.read32(self._base + RTXTargetThread.SP_OFFSET)
        except exceptions.TransferError:
            LOG.debug("Transfer error while reading thread's stack pointer @ 0x%08x", self._base + RTXTargetThread.SP_OFFSET)
            return 0

    def get_stack_frame(self):
        # Get "stack frame" (EXC_RETURN value from LR) saved in thread struct.
        # Note that RTX5 only stores bottom byte - hide that by extending.
        try:
            return self._target_context.read8(self._base + RTXTargetThread.STACKFRAME_OFFSET) | 0xFFFFFF00
        except exceptions.TransferError:
            LOG.debug("Transfer error while reading thread's stack frame @ 0x%08x", self._base + RTXTargetThread.STACKFRAME_OFFSET)
            return 0xFFFFFFFD

class RTX5ThreadProvider(ThreadProvider):
    """@brief Thread provider for RTX5 RTOS."""

    # Offsets in osRtxInfo_t
    KERNEL_STATE_OFFSET = 8
    CURRENT_OFFSET = 20
    THREADLIST_OFFSET = 36
    DELAYLIST_OFFSET = 44
    WAITLIST_OFFSET = 48

    # Offset in osRtxThread_t
    THREADNEXT_OFFSET = 8
    DELAYNEXT_OFFSET = 16

    def __init__(self, target):
        super(RTX5ThreadProvider, self).__init__(target)

    def init(self, symbolProvider):
        # Lookup required symbols.
        self._os_rtx_info = symbolProvider.get_symbol_value('osRtxInfo')
        if self._os_rtx_info is None:
            return False
        LOG.debug('osRtxInfo = 0x%08x', self._os_rtx_info)
        self._readylist = self._os_rtx_info + RTX5ThreadProvider.THREADLIST_OFFSET
        self._delaylist = self._os_rtx_info + RTX5ThreadProvider.DELAYLIST_OFFSET
        self._waitlist = self._os_rtx_info + RTX5ThreadProvider.WAITLIST_OFFSET
        self._threads = {}
        self._current = None
        self._current_id = None
        self._target.session.subscribe(self.event_handler, Target.Event.POST_FLASH_PROGRAM)
        self._target.session.subscribe(self.event_handler, Target.Event.POST_RESET)
        return True

    def get_threads(self):
        if not self.is_enabled:
            return []
        return list(self._threads.values())

    def invalidate(self):
        self._threads = {}

    def event_handler(self, notification):
        # Invalidate threads list if flash is reprogrammed.
        self.invalidate();

    def _build_thread_list(self):
        newThreads = {}

        def create_or_update(thread):
            # Check for and reuse existing thread.
            if thread in self._threads:
                # Thread already exists, update its state.
                t = self._threads[thread]
                t.update_state()
            else:
                # Create a new thread.
                t = RTXTargetThread(self._target_context, self, thread)
            newThreads[t.unique_id] = t

        # Currently running Thread
        thread = self._target_context.read32(self._os_rtx_info + RTX5ThreadProvider.CURRENT_OFFSET)
        if thread:
            create_or_update(thread)
            self._current_id = thread
            self._current = newThreads[thread]
        else:
            self._current_id = None
            self._current = None

        # List of target thread lists to examine.
        threadLists = [
            TargetList(self._target_context, self._readylist, RTX5ThreadProvider.THREADNEXT_OFFSET),
            TargetList(self._target_context, self._delaylist, RTX5ThreadProvider.DELAYNEXT_OFFSET),
            TargetList(self._target_context, self._waitlist, RTX5ThreadProvider.DELAYNEXT_OFFSET),
            ]

        # Scan thread lists.
        for theList in threadLists:
            for thread in theList:
                create_or_update(thread)

        # Create fake handler mode thread.
        if self._target_context.read_core_register('ipsr') > 0:
            newThreads[HandlerModeThread.UNIQUE_ID] = HandlerModeThread(self._target_context, self)

        self._threads = newThreads

    def get_thread(self, threadId):
        if not self.is_enabled:
            return None
        self.update_threads()
        return self._threads.get(threadId, None)

    @property
    def is_enabled(self):
        if self._os_rtx_info is None:
            return False
        try:
            # If we're in Thread mode on the main stack, can't be active, even
            # if kernel state says we are (eg post reset)
            return self.get_kernel_state() != 0 and not self._target.in_thread_mode_on_main_stack()
        except exceptions.TransferError as exc:
            LOG.debug("Transfer error checking if enabled: %s", exc)
            return False

    @property
    def current_thread(self):
        if not self.is_enabled:
            return None
        self.update_threads()
        id = self.get_current_thread_id()
        try:
            return self._threads[id]
        except KeyError:
            LOG.debug("key error getting current thread id=%s; self._threads = %s",
                ("%x" % id) if (id is not None) else id, repr(self._threads))
            return None

    def is_valid_thread_id(self, threadId):
        if not self.is_enabled:
            return False
        self.update_threads()
        return threadId in self._threads

    def get_current_thread_id(self):
        if not self.is_enabled:
            return None
        if self._target_context.read_core_register('ipsr') > 0:
            return HandlerModeThread.UNIQUE_ID
        return self.get_actual_current_thread_id()

    def get_actual_current_thread_id(self):
        if not self.is_enabled:
            return None
        self.update_threads()
        return self._current_id

    def get_kernel_state(self):
        return self._target_context.read8(self._os_rtx_info + RTX5ThreadProvider.KERNEL_STATE_OFFSET)

class RTX5Plugin(Plugin):
    """@brief Plugin class for the RTX5 RTOS."""

    def load(self):
        return RTX5ThreadProvider

    @property
    def name(self):
        return "rtx5"

    @property
    def description(self):
        return "RTX5"



# pyOCD debugger
# Copyright (c) 2023 David van Rijn
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import logging
from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...coresight.cortex_m import CortexM
from ...coresight.minimal_mem_ap import MinimalMemAP as MiniAP

LOG = logging.getLogger(__name__)

class DBGMCU:

    # Via APB-ap (AP2)
    #BASE =   0xe00e1000
    # via AHB-ap (AP0,1)
    BASE =   0x5c001000

    IDC =  BASE + 0x000
    CR =   BASE + 0x004
    CR_VALUE = (0x3f | # keep running in stop sleep and standby
               0x07 << 20 | # enable all debug components
               0x07
               )

    ABP3 = BASE + 0x034

class FlashPeripheral:
    def __init__(self, bank=0):
        assert bank < 2, "only two banks on this device"

        # only per-bank registers are offset
        offset = 0x100 if bank == 1 else 0
        self.bank = bank
        self.flashaddr = 0x2000+0x12000000+0x40000000
        self.flash_keyr      = self.flashaddr + 0x04 + offset
        self.flash_optkeyr   = self.flashaddr + 0x08
        self.flash_optcr     = self.flashaddr + 0x18
        self.flash_cr        = self.flashaddr + 0x0c + offset
        self.flash_sr        = self.flashaddr + 0x10 + offset
        self.flash_optsr_cur = self.flashaddr + 0x1c + offset
        self.flash_optsr_prg = self.flashaddr + 0x20 + offset



FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xe7fdbe00,
    0x8f4ff3bf, 0xf64a4770, 0x49fe20aa, 0x10406008, 0x20066008, 0x60081d09, 0x20aaf64a, 0x600849fa,
    0x60081040, 0x1d092006, 0xf2406008, 0x49f710ff, 0x207f6008, 0x60081f09, 0xb5104770, 0x460c4603,
    0xf44fb672, 0x49f27080, 0x20076008, 0x600849f1, 0x00def44f, 0xbf006148, 0x690048ee, 0x0004f000,
    0xd1f92800, 0x68c048eb, 0x0001f000, 0x48eab120, 0x604849e8, 0x604848e9, 0x00def44f, 0x600849e8,
    0x48e7bf00, 0x68001f00, 0x0004f000, 0xd1f82800, 0x380848e3, 0xf0006800, 0xb1380001, 0x49e048de,
    0x60083910, 0x49db48dd, 0x0104f8c1, 0xffabf7ff, 0x49d848dc, 0x49da6148, 0x20006008, 0x4601bd10,
    0x47702000, 0x49d348d8, 0xbf006148, 0x690048d1, 0x0004f000, 0xd1f92800, 0x49d148d3, 0xbf006008,
    0x1f0048cf, 0xf0006800, 0x28000004, 0xbf00d1f8, 0x690048c8, 0x0004f000, 0xd1f92800, 0x68c048c5,
    0x0030f020, 0x60c849c3, 0x68c04608, 0x0008f040, 0x460860c8, 0xf04068c0, 0x60c80080, 0x48bdbf00,
    0xf0006900, 0x28000004, 0x48bad1f9, 0xf02068c0, 0x49b80008, 0xbf0060c8, 0x1f0048b9, 0xf0006800,
    0x28000004, 0x48b6d1f8, 0x68003808, 0x0030f020, 0xf8c149b0, 0x4608010c, 0x010cf8d0, 0x0008f040,
    0x010cf8c1, 0xf8d04608, 0xf040010c, 0xf8c10080, 0xbf00010c, 0x1f0048aa, 0xf0006800, 0x28000004,
    0x48a7d1f8, 0x68003808, 0x0008f020, 0xf8c149a1, 0x2000010c, 0xb5704770, 0x461a4603, 0x3600f503,
    0xe0922400, 0x4543f3c2, 0x6f00f1b2, 0xf1b2d33f, 0xd23c6f01, 0x69404897, 0x310e499b, 0x49954308,
    0xe0016148, 0xff1ff7ff, 0x69004892, 0x0004f000, 0xd1f72800, 0x68c0488f, 0x7130f647, 0x498d4388,
    0xf04460c8, 0xea400004, 0xf0402005, 0x68c90030, 0x49884308, 0x460860c8, 0xf04068c0, 0x60c80080,
    0xf7ffe001, 0x4883ff00, 0xf0006900, 0x28000004, 0x4880d1f7, 0xf02068c0, 0x497e0004, 0x460860c8,
    0xf0006900, 0x28000001, 0x2001d04c, 0x487cbd70, 0x497d6800, 0x4308310e, 0xf8c14976, 0xe0010114,
    0xfee1f7ff, 0x1f004876, 0xf0006800, 0x28000004, 0x4873d1f6, 0x68003808, 0x7130f647, 0x496d4388,
    0x010cf8c1, 0x0104f044, 0x0008f1a5, 0x2000ea41, 0x0030f040, 0xf8d14967, 0x4308110c, 0xf8c14965,
    0x4608010c, 0x010cf8d0, 0x0080f040, 0x010cf8c1, 0xf7ffe001, 0x4862feb8, 0x68001f00, 0x0004f000,
    0xd1f62800, 0x3808485e, 0xf0206800, 0x49590004, 0x010cf8c1, 0x1f00485a, 0xf0006800, 0xb1080001,
    0xe7b32001, 0x3200f502, 0x42b2bf00, 0xaf6af67f, 0xe7ab2000, 0x4df7e92d, 0x46924605, 0x9c01462f,
    0x46d0463a, 0xf1b72300, 0xd30a6f00, 0x6f01f1b7, 0x4848d207, 0x494c6940, 0x4308310e, 0x61484945,
    0x4847e007, 0x49486800, 0x4308310e, 0xf8c14941, 0xe0ae0114, 0xfe77f7ff, 0x6f00f1b2, 0xf1b2d30a,
    0xd2076f01, 0x68c0483b, 0x7130f647, 0x49394388, 0xe00860c8, 0x3808483a, 0xf6476800, 0x43887130,
    0xf8c14934, 0xf1b2010c, 0xd3066f00, 0x6f01f1b2, 0x2032d203, 0x60c8492f, 0x2032e003, 0x39084930,
    0x2c206008, 0x2300d30f, 0xf8d8e009, 0xf8d81000, 0x60110004, 0xf1086050, 0x32080808, 0x2b041c5b,
    0x3c20dbf3, 0x4616e015, 0x230046c3, 0xf81be004, 0xf8060b01, 0x1c5b0b01, 0xd3f842a3, 0xe0032300,
    0xf80620ff, 0x1c5b0b01, 0x0020f1c4, 0xd8f74298, 0xf7ff2400, 0xf1b2fe25, 0xd30c6f00, 0x6f01f1b2,
    0xe001d209, 0xfe1ff7ff, 0x69004812, 0x0004f000, 0xd1f72800, 0xe001e009, 0xfe15f7ff, 0x1f004810,
    0xf0006800, 0x28000004, 0x480ad1f6, 0x20006900, 0xf1b2b358, 0xd31e6f00, 0x6f01f1b2, 0x4805d21b,
    0xe01368c0, 0x58004800, 0x58004c00, 0x40002c04, 0x580244d4, 0x52002000, 0x45670123, 0xcdef89ab,
    0x52002114, 0x0fee0000, 0x0fef0000, 0x0002f020, 0x60c84945, 0x4845e006, 0xf0206800, 0x49420002,
    0x010cf8c1, 0xe8bd2000, 0xf1b28dfe, 0xd3096f00, 0x6f01f1b2, 0x483cd206, 0xf02068c0, 0x493a0002,
    0xe00660c8, 0x68004839, 0x0002f020, 0xf8c14936, 0x2c00010c, 0xaf4ef47f, 0xe7e42000, 0x68004834,
    0x0001f040, 0x60084932, 0x30104831, 0x49316800, 0x492f4008, 0x60083110, 0x6800482d, 0x4008492e,
    0x6008492b, 0x68004608, 0x2080f420, 0x48286008, 0x68003010, 0x00fef420, 0x31104925, 0x20006008,
    0x31604923, 0x48226008, 0xf0206800, 0x49200018, 0x481f6008, 0x68003010, 0x3110491d, 0x46086008,
    0x60086800, 0x68004608, 0x7000f440, 0x05c86008, 0x6008491a, 0xb5004770, 0xf7ff2200, 0x4812fd84,
    0xe0016902, 0x69024810, 0x0001f002, 0xd1f92800, 0x1d00480e, 0xe0026802, 0x1d00480c, 0xf0026802,
    0x28000001, 0x4808d1f8, 0xe00168c2, 0x68c24806, 0x28002000, 0x4805d1fa, 0xe0016802, 0x68024803,
    0x28002000, 0xbd00d1fa, 0x52002000, 0x5200210c, 0x58024400, 0xf87fc00c, 0xfef6ffff, 0xe000ed08,
    0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x2000003f,
    'pc_unInit': 0x200000c3,
    'pc_program_page': 0x200002d9,
    'pc_erase_sector': 0x2000019b,
    'pc_eraseAll': 0x200000c9,

    'static_base' : 0x20000000 + 0x00000004 + 0x00000560,
    'begin_stack' : 0x20001d70,
    'end_stack' : 0x20000d70,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x400,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    # Enable double buffering
    'page_buffers' : [
        0x20000570,
        0x20000970
    ],
    'min_program_length' : 0x400,

    # Relative region addresses and sizes
    'ro_start': 0x4,
    'ro_size': 0x560,
    'rw_start': 0x564,
    'rw_size': 0x4,
    'zi_start': 0x568,
    'zi_size': 0x0,

    # Flash information
    'flash_start': 0x8000000,
    'flash_size': 0x200000,
    'sector_sizes': (
        (0x0, 0x2000),
    )
}


class STM32H743xx(CoreSightTarget):

    VENDOR = "STMicroelectronics"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x0800_0000, length=0x10_0000, sector_size=0x8000,
                                                    page_size=0x400,
                                                    is_boot_memory=True,
                                                    algo=FLASH_ALGO),

        FlashRegion( start=0x0810_0000, length=0x10_0000, sector_size=0x8000,
                                                        page_size=0x400,
                                                        algo=FLASH_ALGO),
        #ITCM
        RamRegion(   start=0x00000000, length=0x10000,
                  is_cachable=False,
                  access="rwx"),
        #DTCM
        RamRegion(   start=0x20000000, length=0x20000,
                  is_cachable=False,
                  access="rw"),
        #sram1
        RamRegion(   start=0x30000000, length=0x20000,
                  is_powered_on_boot=False),
        #sram2
        RamRegion(   start=0x30020000, length=0x20000,
                  is_powered_on_boot=False),

        #sram3
        RamRegion(   start=0x30040000, length=0x8000,
                  is_powered_on_boot=False),
        #sram4
        RamRegion(   start=0x38000000, length=0x10000),
        )

    def __init__(self, session):
        super().__init__(session, self.MEMORY_MAP)

    def assert_reset_for_connect(self):
        self.dp.assert_reset(1)

    def safe_reset_and_halt(self):
        assert self.dp.is_reset_asserted()

        # At this point we can't access full AP as it is not initialized yet.
        # Let's create a minimalistic AP and use it.
        ap = MiniAP(self.dp)
        ap.init()

        DEMCR_value = ap.read32(CortexM.DEMCR)

        # Halt on reset.
        ap.write32(CortexM.DEMCR,
                   CortexM.DEMCR_VC_CORERESET |
                   CortexM.DEMCR_TRCENA
                   )
        ap.write32(CortexM.DHCSR, CortexM.DBGKEY | CortexM.C_DEBUGEN)

        self.dp.assert_reset(0)
        time.sleep(0.01)

        DEV_ID = ap.read32(DBGMCU.IDC) & 0xfff
        assert DEV_ID == 0x450, f"IDC.DEV_ID 0x{DEV_ID:03x} did not match expected. 0x450"
        ap.write32(DBGMCU.CR, DBGMCU.CR_VALUE)

        CR = ap.read32(DBGMCU.CR)
        LOG.info("CR: 0x%08x", CR)

        # Restore DEMCR original value.
        ap.write32(CortexM.DEMCR, DEMCR_value)

    def create_init_sequence(self):
        # this was copied from target_STM32F767xx.py but seems to apply here as well
        #
        # STM32 under some low power/broken clock states doesn't allow AHP communication.
        # Low power modes are quite popular on stm32 (including MBed OS defaults).
        # 'attach' mode is broken by default, as STM32 can't be connected on low-power mode
        #  successfully without previous DBGMCU setup (It is not possible to write DBGMCU).
        # It is also not possible to run full pyOCD discovery code under-reset.
        #
        # As a solution we can setup DBGMCU under reset, halt core and release reset.
        # Unfortunately this code has to be executed _before_ discovery stage
        # and without discovery stage we don't have access to AP/Core.
        # As a solution we can create minimalistic AP implementation and use it
        # to setup core halt.
        # So the sequence for 'halt' connect mode will look like
        # -> Assert reset
        # -> Connect DebugPort
        # -> Setup MiniAp
        # -> Setup halt on reset
        # -> Enable support for debugging in low-power modes
        # -> Release reset
        # -> [Core is halted and reset is released]
        # -> Continue [discovery, create cores, etc]
        seq = super().create_init_sequence()
        if self.session.options.get('connect_mode') in ('halt', 'under-reset'):
            seq.insert_before('dp_init', ('assert_reset_for_connect', self.assert_reset_for_connect))
            seq.insert_after('dp_init', ('safe_reset_and_halt', self.safe_reset_and_halt))

        return seq

    def _unlock_flash_peripheral(self, flash_banks=[0,1]):

        banks = [FlashPeripheral(n) for n in flash_banks]

        LOG.info('unlocking flash peripheral')
        self.reset_and_halt()

        for bank in banks:
            while self.read32(bank.flash_sr) & 1:
                time.sleep(0.1)

            if self.read32(bank.flash_cr) & 1 != 0:
                self.write32(bank.flash_keyr,    0x4567_0123)
                self.write32(bank.flash_keyr,    0xCDEF_89AB)
                while self.read32(bank.flash_sr) & 1:
                    time.sleep(0.1)

        # shared, so only once
        if self.read32(bank.flash_optcr) & 1 != 0:
            self.write32(bank.flash_optkeyr, 0x0819_2A3B)
            self.write32(bank.flash_optkeyr, 0x4C5D_6E7F)


    def is_locked(self, flash_banks=[0,1]):
        banks = [FlashPeripheral(n) for n in flash_banks]

        # return true if either bank
        for bank in banks:
            optsr = self.read32(bank.flash_optsr_prg)
            rdp = optsr & 0x0000_ff00
            if rdp == 0xcc:
                LOG.warning(f"BANK {bank.bank} permanently locked. No unlock possible")
            if rdp != 0xaa:
                return True
        return False

    def disable_read_protection(self, flash_banks=[0,1]):
        self._unlock_flash_peripheral(flash_banks)
        banks = [FlashPeripheral(n) for n in flash_banks]

        for bank in banks:
            while self.read32(bank.flash_sr) & 1:
                time.sleep(0.1)

            optsr = self.read32(bank.flash_optsr_prg)
            self.write32(bank.flash_optsr_prg, optsr & 0xffff_00ff | 0x0000_aa00)

        # on trigger on both changes
        self.write32(bank.flash_optcr, 2)
        while self.read32(bank.flash_sr) & 1:
            time.sleep(0.1)

        self.reset_and_halt()

    def mass_erase(self, flash_banks=[0,1]):
        self._unlock_flash_peripheral(flash_banks)
        banks = [FlashPeripheral(n) for n in flash_banks]

        for bank in banks:
            while self.read32(bank.flash_sr) & 1:
                time.sleep(0.1)

            self.write32(bank.flash_cr, 1<<3 | 3<<4)
            self.write32(bank.flash_cr, 1<<3 | 3<<4 | 1<<7)
            LOG.info("mass_erase banks %i", bank.bank)

        # banks can be erased at the same time
        # so start both,
        # then wait for both
        for bank in banks:
            while self.read32(bank.flash_sr) & 1:
                time.sleep(0.1)
            LOG.info("mass_erase bank %i done", bank.bank)






# pyOCD debugger
# Copyright (c) 2013-2019 Arm Limited
# Copyright (c) 2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import string

from .builtin import BUILTIN_TARGETS

## @brief Dictionary of all targets.
#
# This table starts off with only the builtin targets. At runtime it may be extended with
# additional targets from CMSIS DFPs or other sources.
TARGET = BUILTIN_TARGETS.copy()

## @brief Legal characters in target type names.
#
# Basically, C language identifier characters.
_TARGET_TYPE_NAME_CHARS = string.ascii_letters + string.digits + '_'

def normalise_target_type_name(target_type: str) -> str:
    """@brief Normalise a target type name.

    The output string has all non-ASCII alphanumeric characters replaced with underscores and is
    converted to all lowercase. Only one underscore in a row will be inserted in the output. For example,
    "foo--bar" will be normalised to "foo_bar".
    """
    result = ""
    in_replace = False
    for c in target_type:
        if c in _TARGET_TYPE_NAME_CHARS:
            result += c.lower()
            in_replace = False
        elif not in_replace:
            result += '_'
            in_replace = True
    return result



# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2020 Patrick Huesmann
# Copyright (c) 2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import (Any, Dict, List, NamedTuple, Tuple, Union)

class OptionInfo(NamedTuple):
    # TODO Change 'type' field's type Any, and use Union for multi-typed options instead of a tuple of types.
    name: str
    type: Union[type, Tuple[type, ...]]
    default: Any
    help: str

## @brief Definitions of the builtin options.
BUILTIN_OPTIONS = [
    # Common options
    OptionInfo('adi.v5.max_invalid_ap_count', int, 3,
        "If this number of invalid APs is found in a row, then AP scanning will stop. The 'scan_all_aps' option "
        "takes precedence over this option if set."),
    OptionInfo('allow_no_cores', bool, False,
        "Prevents raising an error if no cores were found after CoreSight discovery."),
    OptionInfo('auto_unlock', bool, True,
        "Whether to unlock secured target by erasing."),
    OptionInfo('cache.enable_memory', bool, True,
        "Enable the memory read cache. Default is enabled."),
    OptionInfo('cache.enable_register', bool, True,
        "Enable the core register cache. Default is enabled."),
    OptionInfo('cache.read_code_from_elf', bool, True,
        "Controls whether reads of code sections will be taken from an attached ELF file instead of the "
        "target memory."),
    OptionInfo('chip_erase', str, "sector",
        "Whether to perform a chip erase or sector erases when programming flash. The value must be"
        " one of \"auto\", \"sector\", or \"chip\"."),
    OptionInfo('cmsis_dap.prefer_v1', bool, False,
        "If a device provides both CMSIS-DAP v1 and v2 interfaces, use the v1 interface in preference of v2. "
        "Normal behaviour is to prefer the v2 interface. This option is primarily intended for testing."),
    OptionInfo('commander.history_length', int, 1000,
        "Number of entries in the pyOCD Commander command history. Set to -1 for unlimited. Default is 1000."),
    OptionInfo('config_file', str, None,
        "Path to custom config file."),
    OptionInfo('connect_mode', str, "halt",
        "One of 'halt', 'pre-reset', 'under-reset', 'attach'. Default is 'halt'."),
    OptionInfo('cpu.step.instruction.timeout', float, 0.0,
        "Timeout in seconds for instruction step operations. Defaults to 0, or no timeout."),
    OptionInfo('dap_protocol', str, 'default',
        "Wire protocol, either 'swd', 'jtag', or 'default'."),
    OptionInfo('dap_swj_enable', bool, True,
        "Send SWJ transition sequence to switch between SWD and JTAG."),
    OptionInfo('dap_swj_use_dormant', bool, False,
        "When switching between SWD and JTAG, use the SWJ sequence from ADIv5.2 that utilizes a new dormant state."),
    OptionInfo('debug.log_flm_info', bool, False,
        "Log details of loaded .FLM flash algos."),
    OptionInfo('debug.traceback', bool, False,
        "Print tracebacks for exceptions."),
    OptionInfo('enable_multicore_debug', bool, False,
        "Whether to put pyOCD into multicore debug mode. Doing so changes the default software reset type of "
        "secondary cores to VECTRESET, or emulated reset if that is not supported (i.e., non-v7-M cores)."),
    OptionInfo('fast_program', bool, False,
        "Setting this option to True will use CRC checks of existing flash sector contents to "
        "determine whether pages need to be programmed."),
    OptionInfo('flash.timeout.init', float, 5.0,
        "Flash algorithm init and uninit timeout in seconds."),
    OptionInfo('flash.timeout.analyzer', float, 30.0,
        "Flash algorithm CRC analyzer timeout in seconds."),
    OptionInfo('flash.timeout.erase_all', float, 240.0,
        "Flash algorithm erase all timeout in seconds."),
    OptionInfo('flash.timeout.erase_sector', float, 10.0,
        "Flash algorithm sector erase timeout in seconds."),
    OptionInfo('flash.timeout.program', float, 10.0,
        "Flash algorithm programming timeout in seconds."),
    OptionInfo('frequency', int, 1000000,
        "SWD/JTAG frequency in Hertz."),
    OptionInfo('hide_programming_progress', bool, False,
        "Disables flash programming progress bar."),
    OptionInfo('keep_unwritten', bool, False,
        "Whether to preserve existing flash content for ranges of sectors that will be erased but not "
        "written with new data. Default is False."),
    OptionInfo('logging', (str, dict), None,
        "Logging configuration dictionary, or path to YAML file containing logging configuration."),
    OptionInfo('no_config', bool, False,
        "Do not use default config file."),
    OptionInfo('pack', (str, list), None,
        "Path or list of paths to CMSIS Device Family Packs. Devices defined in the pack(s) are "
        "added to the list of available targets."),
    OptionInfo('pack.debug_sequences.debugvars', str, None,
        "Variable definition statements to change configurable debug sequence variables."),
    OptionInfo('pack.debug_sequences.disabled_sequences', (str, list), None,
        "Comma-separated list of names of debug sequences to disable for a CMSIS-Pack based target. "
        "Disabled sequences can be restricted to a given core by appending a colon and processor "
        "name to the sequence's name. Only top-level debug sequences can be disabled. "
        "Ignored for builtin targets."),
    OptionInfo('pack.debug_sequences.enable', bool, True,
        "Global enable for debug sequences for CMSIS-Pack based targets. Ignored for builtin targets."),
    OptionInfo('primary_core', int, 0,
        "Core number for the primary/boot core of an asymmetric multicore target. This is the core that "
        "will control system reset when 'enable_multicore' is set."),
    OptionInfo('probeserver.port', int, 5555,
        "TCP port for the debug probe server."),
    OptionInfo('project_dir', str, None,
        "Path to the session's project directory. Defaults to the working directory when the pyocd "
        "tool was executed."),
    OptionInfo('reset_type', str, 'default',
        "Which type of reset to use by default ('default', 'hw', 'sw', 'sw_system', 'sw_core', "
        "'sw_sysresetreq', 'sw_vectreset', 'sw_emulated', 'system', 'core', 'sysresetreq', 'vectreset', "
        "'emulated'). The default is 'sw', which itself defaults to 'sw_system'."),
    OptionInfo('reset.hold_time', float, 0.1,
        "Number of seconds to hold hardware reset asserted. Default is 0.1 s (100 ms)."),
    OptionInfo('reset.post_delay', float, 0.1,
        "Number of seconds to delay after a reset is issued. Default is 0.1 s (100 ms)."),
    OptionInfo('reset.halt_timeout', float, 2.0,
        "Timeout for waiting for the core to halt after a reset and halt. Default is 2.0 s."),
    OptionInfo('reset.dap_recover.timeout', float, 2.0,
        "Timeout for waiting for the DAP to be accessible after reset. If the timeout lapses, an attempt will be "
        "made to reconnect the DP and retry. Default is 2.0 s."),
    OptionInfo('reset.core_recover.timeout', float, 2.0,
        "Timeout for waiting for a core to be accessible after reset. A warning is printed if the timeout lapses. "
        "Set to 0 to disable the core accessibility test. Default is 2.0 s."),
    OptionInfo('resume_on_disconnect', bool, True,
        "Whether to run target on disconnect."),
    OptionInfo('scan_all_aps', bool, False,
        "Controls whether all 256 ADIv5 AP addresses will be probed. Default is False."),
    OptionInfo('serve_local_only', bool, True,
        "When this option is True, the GDB server, probe server, and semihosting telnet, and raw SWV "
        "server are only served on localhost. Set to False to enable remote connections."),
    OptionInfo('smart_flash', bool, True,
        "If set to True, the flash loader will attempt to not program pages whose contents are not "
        "going to change by scanning target flash memory. A value of False will force all pages to "
        "be erased and programmed. Default is True."),
    OptionInfo('target_override', str, None,
        "Name of target to use instead of default."),
    OptionInfo('test_binary', str, None,
        "Name of test firmware binary."),
    OptionInfo('user_script', str, None,
        "Path of the user script file."),
    OptionInfo('warning.cortex_m_default', bool, True,
        "Whether to show the warning about use of the cortex_m target type. Default is True."),

    # GDBServer options
    OptionInfo('enable_semihosting', bool, False,
        "Set to True to handle semihosting requests."),
    OptionInfo('enable_swv', bool, False,
        "Whether to enable SWV printf output over the semihosting console. Requires the "
        "swv_system_clock option to be set. The SWO baud rate can be controlled with the "
        "swv_clock option."),
    OptionInfo('debug.status_fault_retry_timeout', float, 1.0,
        "Duration in seconds that a failed target status check will be retried before an error is raised. "
        "Only applies while the target is running after a resume operation in the debugger and pyOCD is waiting "
        "for it to halt again."),
    OptionInfo('gdbserver_port', int, 3333,
        "Base TCP port for the gdbserver."),
    OptionInfo('persist', bool, False,
        "If True, the GDB server will not exit after GDB disconnects."),
    OptionInfo('report_core_number', bool, False,
        "Whether gdb server should report core number as part of the per-thread information."),
    OptionInfo('rtos.enable', bool, True,
        "Overall enable flag for RTOS aware debugging. By default it's enabled but can be switched off "
        "if necessary."),
    OptionInfo('rtos.name', str, None,
        "Name of the RTOS plugin to use. If not set, all RTOS plugins are given a chance to load."),
    OptionInfo('semihost_console_type', str, 'telnet',
        "If set to \"telnet\" then the semihosting telnet server will be started, otherwise "
        "semihosting will print to the console."),
    OptionInfo('semihost_use_syscalls', bool, False,
        "Whether to use GDB syscalls for semihosting file access operations."),
    OptionInfo('semihost.commandline', str, "",
        "Program command line string, used for the SYS_GET_CMDLINE semihosting request."),
    OptionInfo('step_into_interrupt', bool, False,
        "Enable interrupts when performing step operations."),
    OptionInfo('swv_clock', int, 1000000,
        "Frequency in Hertz of the SWO baud rate. Default is 1 MHz."),
    OptionInfo('swv_system_clock', int, None,
        "Frequency in Hertz of the target's system clock. Used to compute the SWO baud rate "
        "divider. No default."),
    OptionInfo('swv_raw_enable', bool, True,
        "Enable flag for the raw SWV stream server."),
    OptionInfo('swv_raw_port', int, 3443,
        "TCP port number for the raw SWV stream server."),
    OptionInfo('telnet_port', int, 4444,
        "Base TCP port number for the semihosting telnet server."),
    OptionInfo('vector_catch', str, 'h',
        "Enable vector catch sources."),
    OptionInfo('xpsr_control_fields', bool, False,
        "When set to True, XPSR and CONTROL registers will have their respective bitfields defined "
        "for presentation in gdb."),
    ]

## @brief The runtime dictionary of options.
OPTIONS_INFO: Dict[str, OptionInfo] = {}

def add_option_set(options: List[OptionInfo]) -> None:
    """@brief Merge a list of OptionInfo objects into OPTIONS_INFO."""
    OPTIONS_INFO.update({oi.name: oi for oi in options})

# Start with only builtin options.
add_option_set(BUILTIN_OPTIONS)



# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import TYPE_CHECKING, cast

from ..core.core_registers import CoreRegisterInfo

if TYPE_CHECKING:
    from ..core.core_registers import CoreRegisterNameOrNumberType

LOG = logging.getLogger(__name__)

# Program Status Register
APSR_MASK = 0xF80F0000
EPSR_MASK = 0x0700FC00
IPSR_MASK = 0x000001FF

class CortexMCoreRegisterInfo(CoreRegisterInfo):
    """@brief Core register subclass for Cortex-M registers.

    For most registers, the index is the value written to the DCRSR register to read or write the
    core register. Other core registers not directly supported by DCRSR have special index values that
    are interpreted by the helper methods on this class and the core register read/write code in CortexM
    and its subclasses. These artificial register index values and how they are interpreted are documented
    in the register definitions in CoreRegisterGroups.
    """

    ## Map of register name to info.
    _NAME_MAP = {}

    ## Map of register index to info.
    _INDEX_MAP = {}

    @classmethod
    def register_name_to_index(cls, reg: "CoreRegisterNameOrNumberType") -> int:
        """@brief Convert a register name to integer register index.
        @param reg Either a register name or internal register number.
        @return Internal register number.
        @exception KeyError
        """
        if isinstance(reg, str):
            try:
                reg = cls._NAME_MAP[reg.lower()].index
            except KeyError as err:
                raise KeyError('unknown core register name %s' % reg) from err
        return reg

    @classmethod
    def get(cls, reg: "CoreRegisterNameOrNumberType") -> "CortexMCoreRegisterInfo":
        """@brief Return the CoreRegisterInfo instance for a register.
        @param reg Either a register name or internal register number.
        @return CoreRegisterInfo
        @exception KeyError
        """
        return cast(CortexMCoreRegisterInfo, super().get(reg))

    @property
    def is_fpu_register(self) -> bool:
        """@brief Returns true for FPSCR, SP, or DP registers."""
        return self.index == 33 or self.is_float_register

    @property
    def is_cfbp_subregister(self) -> bool:
        """@brief Whether the register is one of those combined into CFBP by the DCSR."""
        return -4 <= self.index <= -1

    @property
    def is_psr_subregister(self) -> bool:
        """@brief Whether the register is a combination of xPSR fields."""
        return 0x100 <= self.index <= 0x107

    @property
    def psr_mask(self) -> int:
        """@brief Generate a PSR mask based on bottom 3 bits of a MRS SYSm value"""
        mask = 0
        if (self.index & 1) != 0:
            mask |= IPSR_MASK
        if (self.index & 2) != 0:
            mask |= EPSR_MASK
        if (self.index & 4) == 0:
            mask |= APSR_MASK
        return mask

class CoreRegisterGroups:
    """@brief Namespace for lists of Cortex-M core register information."""

    _I = CortexMCoreRegisterInfo # Reduce table width.

    # For most registers, the index is the DCRSR register selector value. Those registers not directly
    # supported by the DCRSR have special values that are interpreted by the register read/write methods.

    ## @brief Registers common to all M-profile cores.
    M_PROFILE_COMMON = [
        #  Name         index   bits    type            group       gdbnum  feature
        _I('r0',        0,      32,     'int',          'general',  0,      "org.gnu.gdb.arm.m-profile"),
        _I('r1',        1,      32,     'int',          'general',  1,      "org.gnu.gdb.arm.m-profile"),
        _I('r2',        2,      32,     'int',          'general',  2,      "org.gnu.gdb.arm.m-profile"),
        _I('r3',        3,      32,     'int',          'general',  3,      "org.gnu.gdb.arm.m-profile"),
        _I('r4',        4,      32,     'int',          'general',  4,      "org.gnu.gdb.arm.m-profile"),
        _I('r5',        5,      32,     'int',          'general',  5,      "org.gnu.gdb.arm.m-profile"),
        _I('r6',        6,      32,     'int',          'general',  6,      "org.gnu.gdb.arm.m-profile"),
        _I('r7',        7,      32,     'int',          'general',  7,      "org.gnu.gdb.arm.m-profile"),
        _I('r8',        8,      32,     'int',          'general',  8,      "org.gnu.gdb.arm.m-profile"),
        _I('r9',        9,      32,     'int',          'general',  9,      "org.gnu.gdb.arm.m-profile"),
        _I('r10',       10,     32,     'int',          'general',  10,     "org.gnu.gdb.arm.m-profile"),
        _I('r11',       11,     32,     'int',          'general',  11,     "org.gnu.gdb.arm.m-profile"),
        _I('r12',       12,     32,     'int',          'general',  12,     "org.gnu.gdb.arm.m-profile"),
        _I('sp',        13,     32,     'data_ptr',     'general',  13,     "org.gnu.gdb.arm.m-profile"),
        _I('lr',        14,     32,     'code_ptr',     'general',  14,     "org.gnu.gdb.arm.m-profile"),
        _I('pc',        15,     32,     'code_ptr',     'general',  15,     "org.gnu.gdb.arm.m-profile"),
        _I('msp',       17,     32,     'data_ptr',     'system',   16,     "org.gnu.gdb.arm.m-profile"),
        _I('psp',       18,     32,     'data_ptr',     'system',   17,     "org.gnu.gdb.arm.m-profile"),
        _I('primask',   -1,     32,     'int',          'system',   18,     "org.gnu.gdb.arm.m-profile"),
        _I('xpsr',      16,     32,     'int',          'general',  19,     "org.gnu.gdb.arm.m-profile"),
        _I('control',   -4,     32,     'int',          'system',   20,     "org.gnu.gdb.arm.m-profile"),

        # The CONTROL, FAULTMASK, BASEPRI, and PRIMASK registers are special in that they share the
        # same DCRSR register index and are returned as a single value. In this dict, these registers
        # have negative values to signal to the register read/write functions that special handling
        # is necessary. The values are the byte number containing the register value, plus 1 and then
        # negated. So -1 means a mask of 0xff, -2 is 0xff00, and so on. The actual DCRSR register index
        # for these combined registers has the key of 'cfbp'.
        _I('cfbp',      20,     32,     'int',          'system'),

        # Variants of XPSR.
        #
        # XPSR is always read in its entirety via the debug interface, but we also provide
        # aliases for the submasks APSR, IPSR and EPSR. These are encoded as 0x10000 plus 3 lower bits
        # indicating which parts of the PSR we're interested in - encoded the same way as MRS's SYSm.
        # (Note that 'XPSR' continues to denote the raw 32 bits of the register, as per previous versions,
        # not the union of the three APSR+IPSR+EPSR masks which don't cover the entire register).
        #
        #  Name         index   bits    type            group       gdbnum  gdb_feature
        _I('apsr',      0x100,  32,     'int',          'system'),
        _I('iapsr',     0x101,  32,     'int',          'system'),
        _I('eapsr',     0x102,  32,     'int',          'system'),
        _I('ipsr',      0x105,  32,     'int',          'system'),
        _I('epsr',      0x106,  32,     'int',          'system'),
        _I('iepsr',     0x107,  32,     'int',          'system'),
        ]

    ## @brief Registers available only on v7-M and v8-M.ML.
    V7M_v8M_ML_ONLY = [
        #  Name         index   bits    type            group       gdbnum  feature
        _I('basepri',   -2,     32,     'int',          'system',   38,     "org.gnu.gdb.arm.m-profile"),
        _I('faultmask', -3,     32,     'int',          'system',   39,     "org.gnu.gdb.arm.m-profile"),
        ]

    ## @brief Extra registers available with only with the Security extension.
    V8M_SEC_ONLY = [
        #  Name         index   bits    type            group      gdbnum  feature
        _I('msp_ns',    24,     32,     'data_ptr',     'stack',    40,     "v8-m.sp"),
        _I('psp_ns',    25,     32,     'data_ptr',     'stack',    41,     "v8-m.sp"),
        _I('msp_s',     26,     32,     'data_ptr',     'stack',    42,     "v8-m.sp"),
        _I('psp_s',     27,     32,     'data_ptr',     'stack',    43,     "v8-m.sp"),
        _I('msplim_s',  28,     32,     'int',          'stack',    46,     "v8-m.sp"),
        _I('psplim_s',  29,     32,     'int',          'stack',    47,     "v8-m.sp"),
        _I('cfbp_s',    34,     32,     'int',          'system'),
        _I('cfbp_ns',   35,     32,     'int',          'system'),
        ]

    ## @brief The NS stack limits are only available when both Main and Security extensions are present.
    V8M_ML_SEC_ONLY = [
        #  Name         index   bits    type            group       gdbnum  feature
        _I('msplim_ns', 30,     32,     'int',          'stack',    44,     "v8-m.sp"),
        _I('psplim_ns', 31,     32,     'int',          'stack',    45,     "v8-m.sp"),
        ]

    ## @brief Registers only available with the MVE extension.
    V81M_MVE_ONLY = [
        #  Name         index   bits    type            group       gdbnum  feature
        _I('vpr',       36,     32,     'int',          'mve',      44,     "v8-m.mve"),
        ]

    ## @brief VFPv5 floating point registers.
    #
    # GDB understands the double/single separation so we don't need to separately pass the single regs,
    # just the double regs; thus only the double regs have a gdb regnum and feature.
    VFP_V5 = [
        #  Name         index   bits    type            group       gdbnum  gdb_feature
        _I('fpscr',     33,     32,     'int',          'float',    21,     "org.gnu.gdb.arm.vfp"),

        # Single-precision float registers.
        #
        #  Name         index   bits    type            group       gdbnum  gdb_feature
        _I('s0',        0x40,   32,     'ieee_single',  'float'),
        _I('s1',        0x41,   32,     'ieee_single',  'float'),
        _I('s2',        0x42,   32,     'ieee_single',  'float'),
        _I('s3',        0x43,   32,     'ieee_single',  'float'),
        _I('s4',        0x44,   32,     'ieee_single',  'float'),
        _I('s5',        0x45,   32,     'ieee_single',  'float'),
        _I('s6',        0x46,   32,     'ieee_single',  'float'),
        _I('s7',        0x47,   32,     'ieee_single',  'float'),
        _I('s8',        0x48,   32,     'ieee_single',  'float'),
        _I('s9',        0x49,   32,     'ieee_single',  'float'),
        _I('s10',       0x4a,   32,     'ieee_single',  'float'),
        _I('s11',       0x4b,   32,     'ieee_single',  'float'),
        _I('s12',       0x4c,   32,     'ieee_single',  'float'),
        _I('s13',       0x4d,   32,     'ieee_single',  'float'),
        _I('s14',       0x4e,   32,     'ieee_single',  'float'),
        _I('s15',       0x4f,   32,     'ieee_single',  'float'),
        _I('s16',       0x50,   32,     'ieee_single',  'float'),
        _I('s17',       0x51,   32,     'ieee_single',  'float'),
        _I('s18',       0x52,   32,     'ieee_single',  'float'),
        _I('s19',       0x53,   32,     'ieee_single',  'float'),
        _I('s20',       0x54,   32,     'ieee_single',  'float'),
        _I('s21',       0x55,   32,     'ieee_single',  'float'),
        _I('s22',       0x56,   32,     'ieee_single',  'float'),
        _I('s23',       0x57,   32,     'ieee_single',  'float'),
        _I('s24',       0x58,   32,     'ieee_single',  'float'),
        _I('s25',       0x59,   32,     'ieee_single',  'float'),
        _I('s26',       0x5a,   32,     'ieee_single',  'float'),
        _I('s27',       0x5b,   32,     'ieee_single',  'float'),
        _I('s28',       0x5c,   32,     'ieee_single',  'float'),
        _I('s29',       0x5d,   32,     'ieee_single',  'float'),
        _I('s30',       0x5e,   32,     'ieee_single',  'float'),
        _I('s31',       0x5f,   32,     'ieee_single',  'float'),

        # Double-precision float registers.
        #
        # The double-precision floating point registers (D0-D15) are composed of two single-precision
        # floating point registers (S0-S31). The index for double-precision registers is the negated
        # value of the first associated single-precision register.
        #
        #  Name         index   bits    type            group       gdbnum  gdb_feature
        _I('d0',        -0x40,  64,     'ieee_double',  'double',   22,     "org.gnu.gdb.arm.vfp"),
        _I('d1',        -0x42,  64,     'ieee_double',  'double',   23,     "org.gnu.gdb.arm.vfp"),
        _I('d2',        -0x44,  64,     'ieee_double',  'double',   24,     "org.gnu.gdb.arm.vfp"),
        _I('d3',        -0x46,  64,     'ieee_double',  'double',   25,     "org.gnu.gdb.arm.vfp"),
        _I('d4',        -0x48,  64,     'ieee_double',  'double',   26,     "org.gnu.gdb.arm.vfp"),
        _I('d5',        -0x4a,  64,     'ieee_double',  'double',   27,     "org.gnu.gdb.arm.vfp"),
        _I('d6',        -0x4c,  64,     'ieee_double',  'double',   28,     "org.gnu.gdb.arm.vfp"),
        _I('d7',        -0x4e,  64,     'ieee_double',  'double',   29,     "org.gnu.gdb.arm.vfp"),
        _I('d8',        -0x50,  64,     'ieee_double',  'double',   30,     "org.gnu.gdb.arm.vfp"),
        _I('d9',        -0x52,  64,     'ieee_double',  'double',   31,     "org.gnu.gdb.arm.vfp"),
        _I('d10',       -0x54,  64,     'ieee_double',  'double',   32,     "org.gnu.gdb.arm.vfp"),
        _I('d11',       -0x56,  64,     'ieee_double',  'double',   33,     "org.gnu.gdb.arm.vfp"),
        _I('d12',       -0x58,  64,     'ieee_double',  'double',   34,     "org.gnu.gdb.arm.vfp"),
        _I('d13',       -0x5a,  64,     'ieee_double',  'double',   35,     "org.gnu.gdb.arm.vfp"),
        _I('d14',       -0x5c,  64,     'ieee_double',  'double',   36,     "org.gnu.gdb.arm.vfp"),
        _I('d15',       -0x5e,  64,     'ieee_double',  'double',   37,     "org.gnu.gdb.arm.vfp"),
        ]

    del _I # Cleanup namespace.

# Build info map.
CortexMCoreRegisterInfo.add_to_map(CoreRegisterGroups.M_PROFILE_COMMON
            + CoreRegisterGroups.V7M_v8M_ML_ONLY
            + CoreRegisterGroups.V8M_SEC_ONLY
            + CoreRegisterGroups.V8M_ML_SEC_ONLY
            + CoreRegisterGroups.V81M_MVE_ONLY
            + CoreRegisterGroups.VFP_V5)

def index_for_reg(name: str) -> int:
    """@brief Utility to easily convert register name to index."""
    return CortexMCoreRegisterInfo.get(name).index



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import List
import logging
from time import sleep

from .base import SubcommandBase
from ..core.helpers import ConnectHelper
from ..core.session import Session
from ..utility.cmdline import convert_session_options
from ..probe.tcp_probe_server import DebugProbeServer

LOG = logging.getLogger(__name__)

class ServerSubcommand(SubcommandBase):
    """@brief `pyocd server` subcommand."""

    NAMES = ['server']
    HELP = "Run debug probe server."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""
        server_parser = argparse.ArgumentParser(description='server', add_help=False)

        server_config_options = server_parser.add_argument_group('configuration')
        server_config_options.add_argument('-j', '--project', '--dir', metavar="PATH", dest="project_dir",
            help="Set the project directory. Defaults to the directory where pyocd was run.")
        server_config_options.add_argument('--config', metavar="PATH",
            help="Specify YAML configuration file. Default is pyocd.yaml or pyocd.yml.")
        server_config_options.add_argument("--no-config", action="store_true", default=None,
            help="Do not use a configuration file.")
        server_config_options.add_argument('-O', action='append', dest='options', metavar="OPTION=VALUE",
            help="Set named option.")
        server_config_options.add_argument("-da", "--daparg", dest="daparg", nargs='+',
            help="Send setting to DAPAccess layer.")

        server_options = server_parser.add_argument_group('probe server')
        server_options.add_argument("-p", "--port", dest="port_number", type=int, default=None,
            help="Set the server's port number (default 5555).")
        server_options.add_argument("--allow-remote", dest="serve_local_only", default=None, action="store_false",
            help="Allow remote TCP/IP connections (default is no).")
        server_options.add_argument("--local-only", default=False, action="store_true",
            help="Server is local only by default. Use --alow-remote to enable remote "
                 "connections. (Currently ignored.)")
        server_options.add_argument("-u", "--uid", "--probe", dest="unique_id",
            help="Select the debug probe by its full or partial unique ID. Optionally prefixed with "
            "'<probe-type>:' where <probe-type> is the name of a probe plugin.")
        server_options.add_argument("-W", "--no-wait", action="store_true",
            help="Do not wait for a probe to be connected if none are available.")

        return [cls.CommonOptions.LOGGING, server_parser]

    def invoke(self) -> None:
        """@brief Handle 'server' subcommand."""
        # Create a session to load config, particularly logging config. Even though we do have a
        # probe, we don't set it in the session because we don't want the board, target, etc objects
        # to be created.
        session_options = convert_session_options(self._args.options)
        session = Session(probe=None,
                serve_local_only=self._args.serve_local_only,
                options=session_options)

        # The ultimate intent is to serve all available probes by default. For now we just serve
        # a single probe.
        probe = ConnectHelper.choose_probe(unique_id=self._args.unique_id)
        if probe is None:
            return

        # Assign the session to the probe.
        probe.session = session

        # Create the server instance.
        server = DebugProbeServer(session, probe, self._args.port_number, self._args.serve_local_only)
        session.probeserver = server
        LOG.debug("Starting debug probe server")
        server.start()

        # Loop as long as the probe is running. The server thread is a daemon, so the main thread
        # must continue to exist.
        try:
            while server.is_running:
                sleep(0.1)
        except (KeyboardInterrupt, Exception):
            server.stop()
            raise




# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (C) 2020 Ted Tawara
# Copyright (c) 2021 Chris Reed
# Copyright (c) 2021 Matthias Wauer
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import sleep
import logging

from ...utility.sequencer import CallSequence
from ...core import exceptions
from ...core.target import Target
from ...coresight.coresight_target import CoreSightTarget
from ...coresight.cortex_m import CortexM
from ...coresight.cortex_m_v8m import CortexM_v8M
from ...utility import timeout

FPB_CTRL                = 0xE0002000
FPB_COMP0               = 0xE0002008
DWT_COMP0               = 0xE0001020
DWT_FUNCTION0           = 0xE0001028
DWT_FUNCTION_MATCH      = 0x4 << 0   # Instruction address.
DWT_FUNCTION_ACTION     = 0x1 << 4   # Generate debug event.
DWT_FUNCTION_DATAVSIZE  = 0x2 << 10  # 4 bytes.

PERIPHERAL_BASE_NS = 0x40000000
PERIPHERAL_BASE_S  = 0x50000000

FLASH_CMD               = 0x00034000
FLASH_STARTA            = 0x00034010
FLASH_STOPA             = 0x00034014
FLASH_DATAW0            = 0x00034080
FLASH_INT_STATUS        = 0x00034FE0
FLASH_INT_CLR_STATUS    = 0x00034FE8
FLASH_CMD_READ_SINGLE_WORD = 0x3
FLASH_CMD_BLANK_CHECK      = 0x5

BOOTROM_MAGIC_ADDR      = 0x50000040

DM_AP                   = 2

# Control and Status Word (CSW) is used to control the Debug Mailbox communication
DM_CSW = 0x00
# Debugger will set this bit to 1 to request a resynchronrisation
DM_CSW_RESYNCH_REQ_MASK =    (1<<0)
# Write only bit. Once written will cause the chip to reset (note that the DM is not
# reset by this reset as it is only resettable by a SOFT reset or a POR/BOD event)
DM_CSW_CHIP_RESET_REQ_MASK = (1<<5)

# Request register is used to send data from debugger to device
DM_REQUEST = 0x04

# Return register is used to send data from device to debugger
# Note: Any read from debugger side will be stalled until new data is present.
DM_RETURN = 0x08

## Debugger mailbox command to start a debug session (unlock debug).
DM_START_DBG_SESSION = 7

LOG = logging.getLogger(__name__)

class LPC5500Family(CoreSightTarget):

    VENDOR = "NXP"

    # Minimum value for the 'adi.v5.max_invalid_ap_count' option.
    _MIN_INVALID_APS = 3

    def create_init_sequence(self):
        seq = super(LPC5500Family, self).create_init_sequence()

        seq.wrap_task('discovery', self.modify_discovery)
        return seq

    def modify_discovery(self, seq):
        seq.insert_before('find_aps',
                          ('set_max_invalid_aps', self.set_max_invalid_aps)) \
           .insert_before('find_components',
                          ('check_locked_state', lambda : self.check_locked_state(seq))) \
           .wrap_task('find_components', self._modify_ap1) \
           .replace_task('create_cores', self.create_lpc55xx_cores) \
           .insert_before('create_components',
                          ('enable_traceclk', self._enable_traceclk),
                        ) \
           .append(('restore_max_invalid_aps', self.restore_max_invalid_aps))
        return seq

    def set_max_invalid_aps(self):
        # Save current option and make sure it is set to at least 3.
        self._saved_max_invalid_aps = self.session.options.get('adi.v5.max_invalid_ap_count')
        if self._saved_max_invalid_aps < self._MIN_INVALID_APS:
            self.session.options.set('adi.v5.max_invalid_ap_count', self._MIN_INVALID_APS)

    def restore_max_invalid_aps(self):
        # Only restore if we changed it.
        if self._saved_max_invalid_aps < self._MIN_INVALID_APS:
            self.session.options.set('adi.v5.max_invalid_ap_count', self._saved_max_invalid_aps)

    def _modify_ap1(self, seq):
        # If AP#1 exists we need to adjust it before we can read the ROM.
        if seq.has_task('init_ap.1'):
            seq.insert_before('init_ap.1',
                ('set_ap1_nonsec',        self._set_ap1_nonsec),
                )

        return seq

    def check_locked_state(self, seq):
        """@brief Attempt to unlock cores if they are locked (flash is empty etc.)"""
        # The device is not locked if AP#0 was found and is enabled.
        if (0 in self.aps) and self.aps[0].is_enabled:
            return

        # The debugger mailbox should always be present.
        if not DM_AP in self.aps:
            LOG.error("cannot request debug unlock; no debugger mailbox AP was found")
            return

        # Perform the unlock procedure using the debugger mailbox.
        self.unlock(self.aps[DM_AP])

        # Finished, if not called from init sequence
        if seq is None:
            return

        # re-run discovery
        LOG.info("re-running discovery")
        new_seq = CallSequence()
        for entry in seq:
            if entry[0] == 'check_locked_state':
                break
            new_seq.append(entry)
        self.dp.valid_aps = None
        return new_seq

    def _set_ap1_nonsec(self):
        # Make AP#1 transactions non-secure so transfers will succeed.
        self.aps[1].hnonsec = 1

    def create_lpc55xx_cores(self):
        # Make sure AP#0 was detected.
        if (0 not in self.aps) or (not self.aps[0].is_enabled):
            LOG.error("AP#0 was not found, unable to create core 0")
            return

        try:
            # Create core 0 with a custom class.
            core0 = CortexM_LPC5500(self.session, self.aps[0], self.memory_map, 0)
            core0.default_reset_type = self.ResetType.SW_SYSRESETREQ
            self.aps[0].core = core0
            core0.init()
            self.add_core(core0)
        except exceptions.Error as err:
            LOG.error("Error creating core 0: %s", err, exc_info=self.session.log_tracebacks)

        # Create core 1 if the AP is present. It uses the standard Cortex-M core class for v8-M.
        if (1 in self.aps) and (self.aps[0].is_enabled):
            try:
                core1 = CortexM_v8M(self.session, self.aps[1], self.memory_map, 1)
                core1.default_reset_type = self.ResetType.SW_SYSRESETREQ
                self.aps[1].core = core1
                core1.init()
                self.add_core(core1)
            except exceptions.Error as err:
                LOG.error("Error creating core 1: %s", err, exc_info=self.session.log_tracebacks)

    def _enable_traceclk(self):
        # Don't make it worse if no APs were found.
        if (0 not in self.aps) or (not self.aps[0].is_enabled):
            return

        SYSCON_NS_Base_Addr = 0x40000000
        IOCON_NS_Base_Addr  = 0x40001000
        TRACECLKSEL_Addr    = SYSCON_NS_Base_Addr + 0x268
        TRACECLKDIV_Addr    = SYSCON_NS_Base_Addr + 0x308
        AHBCLKCTRLSET0_Addr = IOCON_NS_Base_Addr  + 0x220

        clksel = self.read32(TRACECLKSEL_Addr)  # Read current TRACECLKSEL value
        if clksel > 2:
            self.write32(TRACECLKSEL_Addr, 0x0) # Select Trace divided clock

        clkdiv = self.read32(TRACECLKDIV_Addr) & 0xFF # Read current TRACECLKDIV value, preserve divider but clear rest to enable
        self.write32(TRACECLKDIV_Addr, clkdiv)

        self.write32(AHBCLKCTRLSET0_Addr, (1 << 13)) # Enable IOCON clock

    def trace_start(self):
        # Configure PIO0_10: FUNC - 6, MODE - 0, SLEW - 1, INVERT - 0, DIGMODE - 0, OD - 0
        self.write32(0x40001028, 0x00000046)

        self.call_delegate('trace_start', target=self, mode=0)

        # On a reset when ITM is enabled, TRACECLKDIV/TRACECLKSEL will be reset
        # even though ITM will remain enabled -- which will cause ITM stimulus
        # writes to hang in the target because the FIFO will never appear ready.
        # To prevent this, we explicitly (re)enable traceclk.
        self._enable_traceclk()

    def unlock(self, dm_ap):
        """@brief Unlock Cores. See UM11126 51.6.1 """
        assert self.dp.probe.is_open

        LOG.info("attempting unlock procedure")

        # Set RESYNCH_REQ (0x1) and CHIP_RESET_REQ (0x20) in DM.CSW.
        dm_ap.write_reg(addr=DM_CSW, data=(DM_CSW_RESYNCH_REQ_MASK | DM_CSW_CHIP_RESET_REQ_MASK))
        dm_ap.dp.flush()

        # Wait for reset to complete.
        sleep(0.1)

        # Read CSW to verify the reset happened and the register is cleared.
        retval = dm_ap.read_reg(addr=DM_CSW)
        if retval != 0:
            LOG.error("debugger mailbox failed to reset the device")
            return

        # Write debug unlock request.
        dm_ap.write_reg(addr=DM_REQUEST, data=DM_START_DBG_SESSION)
        dm_ap.dp.flush()

        # Read reply from boot ROM. The return status is the low half-word.
        retval = dm_ap.read_reg(addr=DM_RETURN) & 0xffff
        if retval != 0:
            LOG.error("received error from unlock attempt (%x)", retval)
            return
        return

class CortexM_LPC5500(CortexM_v8M):

    def reset_and_halt(self, reset_type=None):
        """@brief Perform a reset and stop the core on the reset handler. """
        halt_only = False
        catch_mode = 0

        delegateResult = self.call_delegate('set_reset_catch', core=self, reset_type=reset_type)

        # Save CortexM.DEMCR
        demcr = self.read_memory(CortexM.DEMCR)

        # enable the vector catch
        if not delegateResult:
            # This sequence is copied from the NXP LPC55S69_DFP debug sequence.
            reset_vector = 0xFFFFFFFF

            # Clear reset vector catch, make sure the DWT is enabled.
            self.write32(CortexM.DEMCR, (demcr & ~CortexM.DEMCR_VC_CORERESET) | CortexM.DEMCR_TRCENA)

            # If the processor is in Secure state, we have to access the flash controller
            # through the secure alias.
            if self.get_security_state() == Target.SecurityState.SECURE:
                base = PERIPHERAL_BASE_S
            else:
                base = PERIPHERAL_BASE_NS

            #
            # Check to see if the flash is erased
            #
            self.write32(base + FLASH_STARTA, 0x00000000) # Program flash word start address to 0x0
            self.write32(base + FLASH_STOPA, 0x00000000) # Program flash word stop address to 0x0
            self.write32(base + FLASH_INT_CLR_STATUS, 0x0000000F) # Clear Flash controller status
            self.write32(base + FLASH_CMD, FLASH_CMD_BLANK_CHECK) # Check if page is cleared

            # Wait for flash word read to finish.
            with timeout.Timeout(5.0) as t_o:
                while t_o.check():
                    if (self.read32(base + FLASH_INT_STATUS) & 0x00000004) != 0:
                        break
                    sleep(0.01)

            # Check for error reading flash word.
            if (self.read32(base + FLASH_INT_STATUS) & 0xB) == 0:
                LOG.info("required flash area is erased")
                halt_only = True

            # Use the flash programming model to check if the first flash page is readable, since
            # attempted accesses to erased pages result in bus faults. The start and stop address
            # are both set to 0x0 to probe the sector containing the reset vector.
            self.write32(base + FLASH_STARTA, 0x00000000) # Program flash word start address to 0x0
            self.write32(base + FLASH_STOPA, 0x00000000) # Program flash word stop address to 0x0
            self.write_memory_block32(base + FLASH_DATAW0, [0x00000000] * 8) # Prepare for read
            self.write32(base + FLASH_INT_CLR_STATUS, 0x0000000F) # Clear Flash controller status
            if not halt_only:
                self.write32(base + FLASH_CMD, FLASH_CMD_READ_SINGLE_WORD) # Read single flash word

                # Wait for flash word read to finish.
                with timeout.Timeout(5.0) as t_o:
                    while t_o.check():
                        if (self.read32(base + FLASH_INT_STATUS) & 0x00000004) != 0:
                            break
                        sleep(0.01)

                # Check for error reading flash word.
                if (self.read32(base + FLASH_INT_STATUS) & 0xB) == 0:
                    # Read the reset vector address.
                    reset_vector = self.read32(0x00000004)

            # Break on user application reset vector if we have a valid breakpoint address.
            if reset_vector != 0xFFFFFFFF:
                catch_mode = 1
                self.write32(FPB_COMP0, reset_vector|1) # Program FPB Comparator 0 with reset handler address
                self.write32(FPB_CTRL, 0x00000003)    # Enable FPB
            # No valid user application so use watchpoint to break at end of boot ROM. The ROM
            # writes a special address to signal when it's done.
            else:
                catch_mode = 2
                self.write32(DWT_FUNCTION0, 0)
                self.write32(DWT_COMP0, BOOTROM_MAGIC_ADDR)
                self.write32(DWT_FUNCTION0, (DWT_FUNCTION_MATCH | DWT_FUNCTION_ACTION | DWT_FUNCTION_DATAVSIZE))

            # Read DHCSR to clear potentially set DHCSR.S_RESET_ST bit
            self.read32(CortexM.DHCSR)

            if not halt_only:
                self.reset(reset_type)
            else:
                self.halt()

        # wait until the unit resets
        with timeout.Timeout(1.0) as t_o:
            while t_o.check():
                if self.get_state() not in (Target.State.RESET, Target.State.RUNNING):
                    break
                sleep(0.01)

        if self.get_state() != Target.State.HALTED:
            self.halt()
            LOG.warning("had to forcibly halt core after reset")

        # Make sure the thumb bit is set in XPSR in case the reset handler
        # points to an invalid address.
        xpsr = self.read_core_register('xpsr')
        if xpsr is not None and xpsr & self.XPSR_THUMB == 0:
            self.write_core_register('xpsr', xpsr | self.XPSR_THUMB)

        self.call_delegate('clear_reset_catch', core=self, reset_type=reset_type)

        # Clear breakpoint or watchpoint.
        if catch_mode == 1:
            self.write32(FPB_COMP0, 0)
        elif catch_mode == 2:
            self.write32(DWT_COMP0, 0)
            self.write32(DWT_FUNCTION0, 0)

        # restore vector catch setting
        self.write_memory(CortexM.DEMCR, demcr)

    def reset(self, reset_type):
        # unlock debug access after reset
        super(CortexM_LPC5500, self).reset(reset_type)

        self.session.target.check_locked_state(None)



# pyOCD debugger
# Copyright (c) 2006-2013 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)

FLASH_ALGO = {
    'load_address' : 0x10000000,
    'instructions' : [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x4c0fb5f8, 0x25002032, 0x2607444c, 0x490d60a6, 0x60206065, 0x4f0c4449, 0x91004620, 0x696047b8,
    0xd10b2800, 0x203460a6, 0x60206065, 0x60e04807, 0x99004620, 0x696047b8, 0xd0002800, 0xbdf82001,
    0x00000004, 0x00000018, 0x1fff1ff1, 0x00002ee0, 0x4d0fb5f8, 0x444d0b04, 0x490e606c, 0x60ac2032,
    0x60284449, 0x460f4e0c, 0x47b04628, 0x28006968, 0x606cd10b, 0x60ac2034, 0x48086028, 0x463960e8,
    0x47b04628, 0x28006968, 0x2001d000, 0x0000bdf8, 0x00000004, 0x00000018, 0x1fff1ff1, 0x00002ee0,
    0x47700b00, 0x21004807, 0x22016301, 0x63416342, 0x6b416342, 0xd0fc07c9, 0x49036382, 0x60082002,
    0x47702000, 0x40048040, 0x40048000, 0x4614b5f8, 0xd10e0005, 0x68206861, 0x184068e2, 0x188968a1,
    0x69211840, 0x69611840, 0x69a11840, 0x42401840, 0x4e1061e0, 0x444e0b28, 0x60702132, 0x490e6031,
    0x444960b0, 0x46304f0d, 0x47b89100, 0x28006970, 0x6075d10e, 0x60b42033, 0x20ff6030, 0x60f03001,
    0x61304807, 0x99004630, 0x697047b8, 0xd0002800, 0xbdf82001, 0x00000004, 0x00000018, 0x1fff1ff1,
    0x00002ee0, 0x47702000, 0x00000000,
    ],

    'pc_eraseAll' : 0x10000021,
    'pc_init' : 0x100000C5,
    #'pc_UnInit' : 0x10000165,
    'pc_program_page' : 0x100000ED,
    'pc_erase_sector' : 0x10000071,

    'static_base' : 0x10000000 + 0x00000020 + 0x00000148,
    'begin_data' : 0x10000000 + 0x00000A00,
    # Double buffering is not supported since there is not enough ram
    'begin_stack' : 0x10000800,
    'min_program_length' : 256, #1024,
    'analyzer_supported' : False,
}

class LPC11XX_32(CoreSightTarget):

    VENDOR = "NXP"

    MEMORY_MAP = MemoryMap(
        FlashRegion(    start=0,           length=0x8000,       is_boot_memory=True,
                                                                blocksize=4096,
                                                                page_size=256,
                                                                algo=FLASH_ALGO),
        RamRegion(      start=0x10000000,  length=0x1000)
        )

    def __init__(self, session):
        super(LPC11XX_32, self).__init__(session, self.MEMORY_MAP)

    def reset_and_halt(self, reset_type=None, map_to_user=True):
        super(LPC11XX_32, self).reset_and_halt(reset_type)

        # Remap to use flash and set SP and SP accordingly
        if map_to_user:
            self.write_memory(0x40048000, 0x2, 32)
            sp = self.read_memory(0x0)
            pc = self.read_memory(0x4)
            self.write_core_register('sp', sp)
            self.write_core_register('pc', pc)
